import threading
from typing import Callable, Tuple


class AgentJob(threading.Thread):
    """A class that handles multithreading logic.

    Args:
        function (Callable): The function to be executed in a separate thread.
        args (Tuple): The arguments to be passed to the function.
    """

    def __init__(self, function: Callable, args: Tuple):
        threading.Thread.__init__(self)
        self.function = function
        self.args = args

    def run(self) -> None:
        """Runs the function in a separate thread."""
        self.function(*self.args)



from dataclasses import dataclass, field
from typing import List, Optional
from swarms.structs.agent import Agent
from swarms.structs.conversation import Conversation
from swarms.utils.loguru_logger import logger
from swarms.utils.try_except_wrapper import try_except_wrapper
from swarms.structs.base_workflow import BaseWorkflow


@dataclass
class SequentialWorkflow(BaseWorkflow):
    name: str = "Sequential Workflow"
    description: str = None
    objective: str = None
    max_loops: int = 1
    autosave: bool = False
    saved_state_filepath: Optional[str] = "sequential_workflow_state.json"
    restore_state_filepath: Optional[str] = None
    dashboard: bool = False
    agent_pool: List[Agent] = field(default_factory=list)
    # task_pool: List[str] = field(
    #     default_factory=list
    # )  # List to store tasks

    def __post_init__(self):
        super().__init__()
        self.conversation = Conversation(
            time_enabled=True,
            autosave=True,
        )

        # If objective exists then set it
        if self.objective is not None:
            self.conversation.system_prompt = self.objective

    def workflow_bootup(self):
        logger.info(f"{self.name} is activating...")

        for agent in self.agent_pool:
            logger.info(f"Agent {agent.agent_name} Activated")

    @try_except_wrapper
    def add(self, task: str, agent: Agent, *args, **kwargs):
        self.agent_pool.append(agent)
        # self.task_pool.append(
        #     task
        # )  # Store tasks corresponding to each agent

        return self.conversation.add(
            role=agent.agent_name, content=task, *args, **kwargs
        )

    def reset_workflow(self) -> None:
        self.conversation = {}

    @try_except_wrapper
    def run(self):
        if not self.agent_pool:
            raise ValueError("No agents have been added to the workflow.")

        self.workflow_bootup()
        loops = 0
        while loops < self.max_loops:
            previous_output = None  # Initialize to None; will hold the output of the previous agent
            for i, agent in enumerate(self.agent_pool):
                # Fetch the last task specific to this agent from the conversation history
                tasks_for_agent = [
                    msg["content"]
                    for msg in self.conversation.conversation_history
                    if msg["role"] == agent.agent_name
                ]
                task = tasks_for_agent[-1] if tasks_for_agent else None

                if task is None and previous_output is not None:
                    # If no specific task for this agent, use the output from the previous agent
                    task = previous_output

                if task is None:
                    # If no initial task is found, and there's no previous output, log error and skip this agent
                    logger.error(
                        f"No initial task found for agent {agent.agent_name}, and no previous output to use."
                    )
                    continue

                logger.info(
                    f" \n Agent {i+1} ({agent.agent_name}) is executing the task: {task} \n"
                )

                # Space the log

                output = agent.run(task)
                if output is None:
                    logger.error(
                        f"Agent {agent.agent_name} returned None for task: {task}"
                    )
                    raise ValueError(
                        f"Agent {agent.agent_name} returned None."
                    )

                # Update the conversation history with the new output using agent's role
                self.conversation.add(
                    role=agent.agent_name, content=output
                )
                previous_output = output  # Update the previous_output to pass to the next agent

            loops += 1
        return self.conversation.return_history_as_string()



import json
import logging
import threading

from swarms.structs.base_structure import BaseStructure


class ShortTermMemory(BaseStructure):
    """Short term memory.

    Args:
    return_str (bool, optional): _description_. Defaults to True.
    autosave (bool, optional): _description_. Defaults to True.
    *args: _description_
    **kwargs: _description_


    Example:
    >>> from swarms.memory.short_term_memory import ShortTermMemory
    >>> stm = ShortTermMemory()
    >>> stm.add(role="agent", message="Hello world!")
    >>> stm.add(role="agent", message="How are you?")
    >>> stm.add(role="agent", message="I am fine.")
    >>> stm.add(role="agent", message="How are you?")
    >>> stm.add(role="agent", message="I am fine.")


    """

    def __init__(
        self,
        return_str: bool = True,
        autosave: bool = True,
        *args,
        **kwargs,
    ):
        self.return_str = return_str
        self.autosave = autosave
        self.short_term_memory = []
        self.medium_term_memory = []
        self.lock = threading.Lock()

    def add(self, role: str = None, message: str = None, *args, **kwargs):
        """Add a message to the short term memory.

        Args:
            role (str, optional): _description_. Defaults to None.
            message (str, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        try:
            memory = self.short_term_memory.append(
                {"role": role, "message": message}
            )

            return memory
        except Exception as error:
            print(f"Add to short term memory failed: {error}")
            raise error

    def get_short_term(self):
        """Get the short term memory.

        Returns:
            _type_: _description_
        """
        return self.short_term_memory

    def get_medium_term(self):
        """Get the medium term memory.

        Returns:
            _type_: _description_
        """
        return self.medium_term_memory

    def clear_medium_term(self):
        """Clear the medium term memory."""
        self.medium_term_memory = []

    def get_short_term_memory_str(self, *args, **kwargs):
        """Get the short term memory as a string."""
        return str(self.short_term_memory)

    def update_short_term(
        self, index, role: str, message: str, *args, **kwargs
    ):
        """Update the short term memory.

        Args:
            index (_type_): _description_
            role (str): _description_
            message (str): _description_

        """
        self.short_term_memory[index] = {
            "role": role,
            "message": message,
        }

    def clear(self):
        """Clear the short term memory."""
        self.short_term_memory = []

    def search_memory(self, term):
        """Search the memory for a term.

        Args:
            term (_type_): _description_

        Returns:
            _type_: _description_
        """
        results = {"short_term": [], "medium_term": []}
        for i, message in enumerate(self.short_term_memory):
            if term in message["message"]:
                results["short_term"].append((i, message))
        for i, message in enumerate(self.medium_term_memory):
            if term in message["message"]:
                results["medium_term"].append((i, message))
        return results

    def return_shortmemory_as_str(self):
        """Return the memory as a string.

        Returns:
            _type_: _description_
        """
        return str(self.short_term_memory)

    def move_to_medium_term(self, index):
        """Move a message from the short term memory to the medium term memory.

        Args:
            index (_type_): _description_
        """
        message = self.short_term_memory.pop(index)
        self.medium_term_memory.append(message)

    def return_medium_memory_as_str(self):
        """Return the medium term memory as a string.

        Returns:
            _type_: _description_
        """
        return str(self.medium_term_memory)

    def save_to_file(self, filename: str):
        """Save the memory to a file.

        Args:
            filename (str): _description_
        """
        try:
            with self.lock:
                with open(filename, "w") as f:
                    json.dump(
                        {
                            "short_term_memory": (self.short_term_memory),
                            "medium_term_memory": (
                                self.medium_term_memory
                            ),
                        },
                        f,
                    )

                    logging.info(f"Saved memory to {filename}")
        except Exception as error:
            print(f"Error saving memory to {filename}: {error}")

    def load_from_file(self, filename: str, *args, **kwargs):
        """Load the memory from a file.

        Args:
            filename (str): _description_
        """
        try:
            with self.lock:
                with open(filename) as f:
                    data = json.load(f)
                self.short_term_memory = data.get("short_term_memory", [])
                self.medium_term_memory = data.get(
                    "medium_term_memory", []
                )
                logging.info(f"Loaded memory from {filename}")
        except Exception as error:
            print(f"Erorr loading memory from {filename}: {error}")



import logging
from collections import defaultdict
from swarms.utils.loguru_logger import logger
from swarms.structs.agent import Agent
from typing import Sequence, Callable


class AgentRearrange:
    def __init__(
        self,
        agents: Sequence[Agent] = None,
        verbose: bool = False,
        custom_prompt: str = None,
        callbacks: Sequence[Callable] = None,
        *args,
        **kwargs,
    ):
        """
        Initialize the AgentRearrange class.

        Args:
            agents (Sequence[Agent], optional): A sequence of Agent objects. Defaults to None.
            verbose (bool, optional): Whether to enable verbose mode. Defaults to False.
            custom_prompt (str, optional): A custom prompt string. Defaults to None.
            callbacks (Sequence[Callable], optional): A sequence of callback functions. Defaults to None.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        if not all(isinstance(agent, Agent) for agent in agents):
            raise ValueError(
                "All elements must be instances of the Agent class."
            )
        self.agents = agents
        self.verbose = verbose
        self.custom_prompt = custom_prompt
        self.callbacks = callbacks if callbacks is not None else []
        self.flows = defaultdict(list)

    def parse_pattern(self, pattern: str):
        """
        Parse the interaction pattern and setup task flows.

        Args:
            pattern (str): The interaction pattern to parse.

        Returns:
            bool: True if the pattern parsing is successful, False otherwise.
        """
        try:
            for flow in pattern.split(","):
                parts = [part.strip() for part in flow.split("->")]
                if len(parts) != 2:
                    logging.error(
                        f"Invalid flow pattern: {flow}. Each flow"
                        " must have exactly one '->'."
                    )
                    return False

                source_name, destinations_str = parts
                source = self.find_agent_by_name(source_name)
                if source is None:
                    logging.error(f"Source agent {source_name} not found.")
                    return False

                destinations_names = destinations_str.split()
                for dest_name in destinations_names:
                    dest = self.find_agent_by_name(dest_name)
                    if dest is None:
                        logging.error(
                            f"Destination agent {dest_name} not" " found."
                        )
                        return False
                    self.flows[source.agent_name].append(dest.agent_name)
            return True
        except Exception as e:
            logger.error(f"Error: {e}")
            raise e

    def self_find_agen_by_name(self, name: str):
        """
        Find an agent by its name.

        Args:
            name (str): The name of the agent to find.

        Returns:
            Agent: The Agent object if found, None otherwise.
        """
        for agent in self.agents:
            if agent.agent_name == name:
                return agent
        return None

    def __call__(
        self,
        agents: Sequence[Agent] = None,
        pattern: str = None,
        task: str = None,
        **tasks,
    ):
        """
        Execute the task based on the specified pattern.

        Args:
            agents (Sequence[Agent], optional): A sequence of Agent objects. Defaults to None.
            pattern (str, optional): The interaction pattern to follow. Defaults to None.
            task (str, optional): The task to execute. Defaults to None.
            **tasks: Additional tasks specified as keyword arguments.
        """
        try:
            if agents:
                self.flows.clear()  # Reset previous flows
                if not self.parse_pattern(pattern):
                    return  # Pattern parsing failed

                for source, destinations in self.flows.items():
                    for dest in destinations:
                        dest_agent = self.self_find_agen_by_name(dest)
                        task = tasks.get(dest, task)

                        if self.custom_prompt:
                            dest_agent.run(f"{task} {self.custom_prompt}")
                        else:
                            dest_agent.run(f"{task} (from {source})")
            # else:
            #     raise ValueError(
            #         "No agents provided. Please provide agents to"
            #         " execute the task."
            #     )
        except Exception as e:
            logger.error(
                f"Error: {e} try again by providing agents and" " pattern"
            )
            raise e


# # Example usage
# try:
#     agents = [
#         Agent(agent_name=f"b{i}") for i in range(1, 4)
#     ]  # Creating agents b1, b2, b3
#     agents.append(Agent(agent_name="d"))  # Adding agent d
#     rearranger = Rearrange(agents)

#     # Specifying a complex pattern for task execution
#     rearranger.execute("d -> b1 b2 b3, b2 -> b3", "Analyze data")
# except ValueError as e:
#     logging.error(e)



from swarms.agents.agent_wrapper import agent_wrapper
from swarms.agents.base import AbstractAgent
from swarms.agents.omni_modal_agent import OmniModalAgent
from swarms.agents.simple_agent import SimpleAgent
from swarms.agents.stopping_conditions import (
    check_cancelled,
    check_complete,
    check_done,
    check_end,
    check_error,
    check_exit,
    check_failure,
    check_finished,
    check_stopped,
    check_success,
)
from swarms.agents.tool_agent import ToolAgent
from swarms.agents.worker_agent import Worker

__all__ = [
    "AbstractAgent",
    "ToolAgent",
    "SimpleAgent",
    "OmniModalAgent",
    "check_done",
    "check_finished",
    "check_complete",
    "check_success",
    "check_failure",
    "check_error",
    "check_stopped",
    "check_cancelled",
    "check_exit",
    "check_end",
    "Worker",
    "agent_wrapper",
]



import sched
import time
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Callable, Dict, List, Union

from swarms.structs.agent import Agent
from swarms.structs.conversation import Conversation
from swarms.utils.logger import logger


@dataclass
class Task:
    """
    Task class for running a task in a sequential workflow.

    Attributes:
        description (str): Description of the task.
        agent (Union[Callable, Agent]): Agent or callable object to run the task.
        args (List[Any]): Arguments to pass to the agent or callable object.
        kwargs (Dict[str, Any]): Keyword arguments to pass to the agent or callable object.
        result (Any): Result of the task.
        history (List[Any]): History of the task.
        schedule_time (datetime): Time to schedule the task.
        scheduler (sched.scheduler): Scheduler to schedule the task.
        trigger (Callable): Trigger to run the task.
        action (Callable): Action to run the task.
        condition (Callable): Condition to run the task.
        priority (int): Priority of the task.
        dependencies (List[Task]): List of tasks that need to be completed before this task can be executed.

    Methods:
        execute: Execute the task by calling the agent or model with the arguments and keyword arguments.
        handle_scheduled_task: Handles the execution of a scheduled task.
        set_trigger: Sets the trigger for the task.
        set_action: Sets the action for the task.
        set_condition: Sets the condition for the task.
        is_completed: Checks whether the task has been completed.
        add_dependency: Adds a task to the list of dependencies.
        set_priority: Sets the priority of the task.
        check_dependency_completion: Checks whether all the dependencies have been completed.


    Examples:
    >>> from swarms.structs import Task, Agent
    >>> from swarms.models import OpenAIChat
    >>> agent = Agent(llm=OpenAIChat(openai_api_key=""), max_loops=1, dashboard=False)
    >>> task = Task(description="What's the weather in miami", agent=agent)
    >>> task.execute()
    >>> task.result

    """

    agent: Union[Callable, Agent]
    description: str = None
    result: Any = None
    history: List[Any] = field(default_factory=list)
    schedule_time: datetime = None
    scheduler = sched.scheduler(time.time, time.sleep)
    trigger: Callable = None
    action: Callable = None
    condition: Callable = None
    priority: int = 0
    dependencies: List["Task"] = field(default_factory=list)
    args: List[Any] = field(default_factory=list)
    kwargs: Dict[str, Any] = field(default_factory=dict)

    def execute(self, *args, **kwargs):
        """
        Execute the task by calling the agent or model with the arguments and
        keyword arguments. You can add images to the agent by passing the
        path to the image as a keyword argument.


        Examples:
        >>> from swarms.structs import Task, Agent
        >>> from swarms.models import OpenAIChat
        >>> agent = Agent(llm=OpenAIChat(openai_api_key=""), max_loops=1, dashboard=False)
        >>> task = Task(description="What's the weather in miami", agent=agent)
        >>> task.execute()
        >>> task.result

        """
        logger.info(f"[INFO][Task] Executing task: {self.description}")
        task = self.description
        try:
            if isinstance(self.agent, Agent):
                if self.condition is None or self.condition():
                    self.result = self.agent.run(
                        task=task,
                        *args,
                        **kwargs,
                    )
                    self.history.append(self.result)

                    if self.action is not None:
                        self.action()
            else:
                self.result = self.agent.run(*self.args, **self.kwargs)

            self.history.append(self.result)
        except Exception as error:
            logger.error(f"[ERROR][Task] {error}")

    def run(self, *args, **kwargs):
        self.execute(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        self.execute(*args, **kwargs)

    def handle_scheduled_task(self):
        """
        Handles the execution of a scheduled task.

        If the schedule time is not set or has already passed, the task is executed immediately.
        Otherwise, the task is scheduled to be executed at the specified schedule time.
        """
        logger.info("[INFO][Task] Handling scheduled task")
        try:
            if (
                self.schedule_time is None
                or self.schedule_time <= datetime.now()
            ):
                self.execute()

            else:
                delay = (
                    self.schedule_time - datetime.now()
                ).total_seconds()
                self.scheduler.enter(delay, 1, self.execute)
                self.scheduler_run()
        except Exception as error:
            logger.error(f"[ERROR][Task] {error}")

    def set_trigger(self, trigger: Callable):
        """
        Sets the trigger for the task.

        Args:
            trigger (Callable): The trigger to set.
        """
        self.trigger = trigger

    def set_action(self, action: Callable):
        """
        Sets the action for the task.

        Args:
            action (Callable): The action to set.
        """
        self.action = action

    def set_condition(self, condition: Callable):
        """
        Sets the condition for the task.

        Args:
            condition (Callable): The condition to set.
        """
        self.condition = condition

    def is_completed(self):
        """Is the task completed?

        Returns:
            _type_: _description_
        """
        return self.result is not None

    def add_dependency(self, task):
        """Adds a task to the list of dependencies.

        Args:
            task (_type_): _description_
        """
        self.dependencies.append(task)

    def set_priority(self, priority: int):
        """Sets the priority of the task.

        Args:
            priority (int): _description_
        """
        self.priority = priority

    def check_dependency_completion(self):
        """
        Checks whether all the dependencies have been completed.

        Returns:
            bool: True if all the dependencies have been completed, False otherwise.
        """
        logger.info("[INFO][Task] Checking dependency completion")
        try:
            for task in self.dependencies:
                if not task.is_completed():
                    return False
        except Exception as error:
            logger.error(
                f"[ERROR][Task][check_dependency_completion] {error}"
            )

    def context(
        self,
        task: "Task" = None,
        context: List["Task"] = None,
        *args,
        **kwargs,
    ):
        """
        Set the context for the task.

        Args:
            context (str): The context to set.
        """
        # For sequential workflow, sequentially add the context of the previous task in the list
        new_context = Conversation(time_enabled=True, *args, **kwargs)

        if context:
            for task in context:
                description = (
                    task.description
                    if task.description is not None
                    else ""
                )

                result = task.result if task.result is not None else ""

                # Add the context of the task to the conversation
                new_context.add(
                    task.agent.agent_name, f"{description} {result}"
                )

        elif task:
            description = (
                task.description if task.description is not None else ""
            )
            result = task.result if task.result is not None else ""
            new_context.add(
                task.agent.agent_name, f"{description} {result}"
            )

        prompt = new_context.return_history_as_string()

        # Add to history
        return self.history.append(prompt)



from functools import wraps
from multiprocessing import Manager, Pool, cpu_count
from time import sleep
from typing import Sequence

from swarms.structs.agent import Agent
from swarms.structs.base_workflow import BaseWorkflow
from swarms.utils.loguru_logger import logger


# Retry on failure
def retry_on_failure(max_retries: int = 3, delay: int = 5):
    """
    Decorator that retries a function a specified number of times on failure.

    Args:
        max_retries (int): The maximum number of retries (default: 3).
        delay (int): The delay in seconds between retries (default: 5).

    Returns:
        The result of the function if it succeeds within the maximum number of retries,
        otherwise None.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for _ in range(max_retries):
                try:
                    return func(*args, **kwargs)
                except Exception as error:
                    logger.error(
                        f"Error: {str(error)}, retrying in"
                        f" {delay} seconds..."
                    )
                    sleep(delay)
            return None

        return wrapper

    return decorator


class MultiProcessWorkflow(BaseWorkflow):
    """
    Initialize a MultiProcessWorkflow object.

    Args:
        max_workers (int): The maximum number of workers to use for parallel processing.
        autosave (bool): Flag indicating whether to automatically save the workflow.
        tasks (List[Task]): A list of Task objects representing the workflow tasks.
        *args: Additional positional arguments.
        **kwargs: Additional keyword arguments.

    Example:
    >>> from swarms.structs.multi_process_workflow import MultiProcessingWorkflow
    >>> from swarms.structs.task import Task
    >>> from datetime import datetime
    >>> from time import sleep
    >>>
    >>> # Define a simple task
    >>> def simple_task():
    >>>     sleep(1)
    >>>     return datetime.now()
    >>>
    >>> # Create a task object
    >>> task = Task(
    >>>     name="Simple Task",
    >>>     execute=simple_task,
    >>>     priority=1,
    >>> )
    >>>
    >>> # Create a workflow with the task
    >>> workflow = MultiProcessingWorkflow(tasks=[task])
    >>>
    >>> # Run the workflow
    >>> results = workflow.run(task)
    >>>
    >>> # Print the results
    >>> print(results)
    """

    def __init__(
        self,
        max_workers: int = 5,
        autosave: bool = True,
        agents: Sequence[Agent] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.max_workers = max_workers
        self.autosave = autosave
        self.agents = agents

        self.max_workers or cpu_count()

        # Log
        logger.info(
            (
                "Initialized MultiProcessWorkflow with"
                f" {self.max_workers} max workers and autosave set to"
                f" {self.autosave}"
            ),
        )

        # Log the agents
        if self.agents is not None:
            for agent in self.agents:
                logger.info(f"Agent: {agent.agent_name}")

    def execute_task(self, task: str, *args, **kwargs):
        """Execute a task and handle exceptions.

        Args:
            task (Task): The task to execute.
            *args: Additional positional arguments for the task execution.
            **kwargs: Additional keyword arguments for the task execution.

        Returns:
            Any: The result of the task execution.

        """
        try:
            if self.agents is not None:
                # Execute the task
                for agent in self.agents:
                    result = agent.run(task, *args, **kwargs)

            return result

        except Exception as e:
            logger.error(
                (
                    "An error occurred during execution of task"
                    f" {task}: {str(e)}"
                ),
            )
            return None

    def run(self, task: str, *args, **kwargs):
        """Run the workflow.

        Args:
            task (Task): The task to run.
            *args: Additional positional arguments for the task execution.
            **kwargs: Additional keyword arguments for the task execution.

        Returns:
            List[Any]: The results of all executed tasks.

        """
        try:
            results = []
            with Manager() as manager:
                with Pool(
                    processes=self.max_workers, *args, **kwargs
                ) as pool:
                    # Using manager.list() to collect results in a process safe way
                    results_list = manager.list()
                    jobs = [
                        pool.apply_async(
                            self.execute_task,  # Pass the function, not the function call
                            args=(task,)
                            + args,  # Pass the arguments as a tuple
                            kwds=kwargs,  # Pass the keyword arguments as a dictionary
                            callback=results_list.append,
                            timeout=task.timeout,
                        )
                        for agent in self.agent
                    ]

                    # Wait for all jobs to complete
                    for job in jobs:
                        job.get()

                    results = list(results_list)

                return results
        except Exception as error:
            logger.error(f"Error in run: {error}")
            return None



import logging
import os
from typing import Optional

import requests
from dotenv import load_dotenv

from swarms.models.base_llm import BaseLLM

# Load environment variables
load_dotenv()


def together_api_key_env():
    """Get the API key from the environment."""
    return os.getenv("TOGETHER_API_KEY")


class TogetherLLM(BaseLLM):
    """
    GPT-4 Vision API

    This class is a wrapper for the OpenAI API. It is used to run the GPT-4 Vision model.

    Parameters
    ----------
    together_api_key : str
        The OpenAI API key. Defaults to the together_api_key environment variable.
    max_tokens : int
        The maximum number of tokens to generate. Defaults to 300.


    Methods
    -------
    encode_image(img: str)
        Encode image to base64.
    run(task: str, img: str)
        Run the model.
    __call__(task: str, img: str)
        Run the model.

    Examples:
    ---------
    >>> from swarms.models import GPT4VisionAPI
    >>> llm = GPT4VisionAPI()
    >>> task = "What is the color of the object?"
    >>> img = "https://i.imgur.com/2M2ZGwC.jpeg"
    >>> llm.run(task, img)


    """

    def __init__(
        self,
        together_api_key: str = together_api_key_env,
        model_name: str = "mistralai/Mixtral-8x7B-Instruct-v0.1",
        logging_enabled: bool = False,
        max_workers: int = 10,
        max_tokens: str = 300,
        api_endpoint: str = "https://api.together.xyz",
        beautify: bool = False,
        streaming_enabled: Optional[bool] = False,
        meta_prompt: Optional[bool] = False,
        system_prompt: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super(TogetherLLM).__init__(*args, **kwargs)
        self.together_api_key = together_api_key
        self.logging_enabled = logging_enabled
        self.model_name = model_name
        self.max_workers = max_workers
        self.max_tokens = max_tokens
        self.api_endpoint = api_endpoint
        self.beautify = beautify
        self.streaming_enabled = streaming_enabled
        self.meta_prompt = meta_prompt
        self.system_prompt = system_prompt

        if self.logging_enabled:
            logging.basicConfig(level=logging.DEBUG)
        else:
            # Disable debug logs for requests and urllib3
            logging.getLogger("requests").setLevel(logging.WARNING)
            logging.getLogger("urllib3").setLevel(logging.WARNING)

        if self.meta_prompt:
            self.system_prompt = self.meta_prompt_init()

    # Function to handle vision tasks
    def run(self, task: str = None, *args, **kwargs):
        """Run the model."""
        try:
            headers = {
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self.together_api_key}",
            }
            payload = {
                "model": self.model_name,
                "messages": [
                    {
                        "role": "system",
                        "content": [self.system_prompt],
                    },
                    {
                        "role": "user",
                        "content": task,
                    },
                ],
                "max_tokens": self.max_tokens,
                **kwargs,
            }
            response = requests.post(
                self.api_endpoint,
                headers=headers,
                json=payload,
                *args,
                **kwargs,
            )

            out = response.json()
            content = (
                out["choices"][0].get("message", {}).get("content", None)
            )
            if self.streaming_enabled:
                content = self.stream_response(content)

            return content

        except Exception as error:
            print(
                f"Error with the request: {error}, make sure you"
                " double check input types and positions"
            )
            return None



import logging
from collections import defaultdict
from typing import Callable, Sequence
from swarms.structs.agent import Agent
from swarms.structs.base_swarm import BaseSwarm


# Assuming the existence of an appropriate Agent class and logger setup
class AgentRearrange(BaseSwarm):
    def __init__(
        self,
        agents: Sequence[Agent] = None,
        verbose: bool = False,
        custom_prompt: str = None,
        callbacks: Sequence[Callable] = None,
        *args,
        **kwargs,
    ):
        super().__init__()
        if not all(isinstance(agent, Agent) for agent in agents):
            raise ValueError(
                "All elements must be instances of the Agent class."
            )
        self.agents = agents
        self.verbose = verbose
        self.custom_prompt = custom_prompt
        self.callbacks = callbacks if callbacks is not None else []
        self.flows = defaultdict(list)

    def parse_pattern(self, pattern: str):
        """
        Parse the interaction pattern to set up task flows, supporting both sequential
        and concurrent executions within the same pattern.
        """
        try:
            self.flows.clear()  # Ensure flows are reset each time pattern is parsed
            # Split pattern into potentially concurrent flows
            concurrent_flows = pattern.split(",")
            for flow in concurrent_flows:
                # Trim whitespace and identify sequential parts within each concurrent flow
                parts = [part.strip() for part in flow.split("->")]
                if len(parts) > 1:
                    # Link each part sequentially to the next as source -> destination
                    for i in range(len(parts) - 1):
                        source = parts[i]
                        destination = parts[i + 1]
                        # Validate and add each sequential link
                        if source not in [
                            agent.agent_name for agent in self.agents
                        ]:
                            logging.error(
                                f"Source agent {source} not found."
                            )
                            return False
                        if destination not in [
                            agent.agent_name for agent in self.agents
                        ]:
                            logging.error(
                                f"Destination agent {destination} not"
                                " found."
                            )
                            return False
                        self.flows[source].append(destination)
                else:
                    # Handle single agent case if needed
                    self.flows[parts[0]] = []

            return True
        except Exception as e:
            logging.error(f"Error parsing pattern: {e}")
            return False

    def self_find_agent_by_name(self, name: str):
        for agent in self.agents:
            if agent.agent_name == name:
                return agent
        return None

    def agent_exists(self, name: str):
        for agent in self.agents:
            if agent.agent_name == name:
                return True

        return False

    def parse_concurrent_flow(
        self,
        flow: str,
    ):
        sequential_agents = flow.split("->")
        for i, source_name in enumerate(sequential_agents[:-1]):
            destination_name = sequential_agents[i + 1].strip()
            self.parse_sequential_flow(
                source_name.strip(), destination_name
            )

    def parse_sequential_flow(
        self,
        source: str,
        destination: str,
    ):
        if not self.self_find_agent_by_name(
            source
        ) or not self.self_find_agent_by_name(destination):
            return False
        self.flows[source].append(destination)

    def execute_task(
        self,
        dest_agent_name: str,
        source: str,
        task: str,
        specific_tasks: dict,
    ):
        dest_agent = self.self_find_agent_by_name(dest_agent_name)
        if not dest_agent:
            return None
        task_to_run = specific_tasks.get(dest_agent_name, task)
        if self.custom_prompt:
            out = dest_agent.run(f"{task_to_run} {self.custom_prompt}")
        else:
            out = dest_agent.run(f"{task_to_run} (from {source})")
        return out

    def process_flows(self, pattern, default_task, specific_tasks):
        if not self.parse_pattern(pattern):
            return None

        results = []
        for source, destinations in self.flows.items():
            if not destinations:
                task = specific_tasks.get(source, default_task)
                source_agent = self.self_find_agent_by_name(source)
                if source_agent:
                    result = source_agent.run(task)
                    results.append(result)
            else:
                for destination in destinations:
                    task = specific_tasks.get(destination, default_task)
                    destination_agent = self.self_find_agent_by_name(
                        destination
                    )
                    if destination_agent:
                        result = destination_agent.run(task)
                        results.append(result)
        return results

    def __call__(
        self,
        pattern: str = None,
        default_task: str = None,
        **specific_tasks,
    ):
        self.flows.clear()  # Reset previous flows
        results = self.process_flows(pattern, default_task, specific_tasks)
        return results


# ## Initialize the workflow
# agent = Agent(
#     agent_name="t",
#     agent_description=(
#         "Generate a transcript for a youtube video on what swarms"
#         " are!"
#     ),
#     system_prompt=(
#         "Generate a transcript for a youtube video on what swarms"
#         " are!"
#     ),
#     llm=AnthropicChat(),
#     max_loops=1,
#     autosave=True,
#     dashboard=False,
#     streaming_on=True,
#     verbose=True,
#     stopping_token="<DONE>",
# )

# agent2 = Agent(
#     agent_name="t1",
#     agent_description=(
#         "Generate a transcript for a youtube video on what swarms"
#         " are!"
#     ),
#     llm=AnthropicChat(),
#     max_loops=1,
#     system_prompt="Summarize the transcript",
#     autosave=True,
#     dashboard=False,
#     streaming_on=True,
#     verbose=True,
#     stopping_token="<DONE>",
# )

# agent3 = Agent(
#     agent_name="t2",
#     agent_description=(
#         "Generate a transcript for a youtube video on what swarms"
#         " are!"
#     ),
#     llm=AnthropicChat(),
#     max_loops=1,
#     system_prompt="Finalize the transcript",
#     autosave=True,
#     dashboard=False,
#     streaming_on=True,
#     verbose=True,
#     stopping_token="<DONE>",
# )


# # Rearrange the agents
# rearrange = AgentRearrange(
#     agents=[agent, agent2, agent3],
#     verbose=True,
#     # custom_prompt="Summarize the transcript",
# )

# # Run the workflow on a task
# results = rearrange(
#     # pattern="t -> t1, t2 -> t2",
#     pattern="t -> t1 -> t2",
#     default_task=(
#         "Generate a transcript for a YouTube video on what swarms"
#         " are!"
#     ),
#     t="Generate a transcript for a YouTube video on what swarms are!",
#     # t2="Summarize the transcript",
#     # t3="Finalize the transcript",
# )
# # print(results)



VISION_PROMPT = """
You are a Self-Operating Computer. You use the same operating system as a human.

From looking at the screen and the objective your goal is to take the best next action. 

To operate the computer you have the four options below. 

1. CLICK - Move mouse and click
2. TYPE - Type on the keyboard
3. SEARCH - Search for a program on Mac and open it
4. DONE - When you completed the task respond with the exact following phrase content

Here are the response formats below. 

1. CLICK
Response: CLICK {{ "x": "percent", "y": "percent", "description": "~description here~", "reason": "~reason here~" }} 

2. TYPE
Response: TYPE "value you want to type"

2. SEARCH
Response: SEARCH "app you want to search for on Mac"

3. DONE
Response: DONE

Here are examples of how to respond.
__
Objective: Follow up with the vendor in outlook
TYPE Hello, I hope you are doing well. I wanted to follow up
__
Objective: Open Spotify and play the beatles
SEARCH Spotify
__
Objective: Find a image of a banana
CLICK {{ "x": "50%", "y": "60%", "description": "Click: Google Search field", "reason": "This will allow me to search for a banana" }} 
__
Objective: Go buy a book about the history of the internet
TYPE https://www.amazon.com/
__

A few important notes: 

- Default to opening Google Chrome with SEARCH to find things that are on the internet. 
- Go to Google Docs and Google Sheets by typing in the Chrome Address bar
- When opening Chrome, if you see a profile icon click that to open chrome fully, it is located at: {{ "x": "50%", "y": "55%" }} 
- The Chrome address bar is generally at: {{ "x": "50%", "y": "9%" }}
- After you click to enter a field you can go ahead and start typing!

{previous_action}

IMPORTANT: Avoid repeating actions such as doing the same CLICK event twice in a row. 

Objective: {objective}
"""

USER_QUESTION = (
    "Hello, I can help you with anything. What would you like done?"
)

SUMMARY_PROMPT = """
You are a Self-Operating Computer. You just completed a request from a user by operating the computer. Now you need to share the results. 

You have three pieces of key context about the completed request.

1. The original objective
2. The steps you took to reach the objective that are available in the previous messages
3. The screenshot you are looking at.

Now you need to summarize what you did to reach the objective. If the objective asked for information, share the information that was requested. IMPORTANT: Don't forget to answer a user's question if they asked one.

Thing to note: The user can not respond to your summary. You are just sharing the results of your work.

The original objective was: {objective}

Now share the results!
"""


def format_summary_prompt(objective):
    """
    Format the summary prompt
    """
    prompt = SUMMARY_PROMPT.format(objective=objective)
    return prompt


def format_vision_prompt(objective, previous_action):
    """
    Format the vision prompt
    """
    if previous_action:
        previous_action = (
            "Here was the previous action you took:" f" {previous_action}"
        )
    else:
        previous_action = ""
    prompt = VISION_PROMPT.format(
        objective=objective, previous_action=previous_action
    )
    return prompt



def generate_agent_role_prompt(agent):
    """Generates the agent role prompt.
    Args: agent (str): The type of the agent.
    Returns: str: The agent role prompt.
    """
    prompts = {
        "Finance Agent": (
            "You are a seasoned finance analyst AI assistant. Your"
            " primary goal is to compose comprehensive, astute,"
            " impartial, and methodically arranged financial reports"
            " based on provided data and trends."
        ),
        "Travel Agent": (
            "You are a world-travelled AI tour guide assistant. Your"
            " main purpose is to draft engaging, insightful,"
            " unbiased, and well-structured travel reports on given"
            " locations, including history, attractions, and cultural"
            " insights."
        ),
        "Academic Research Agent": (
            "You are an AI academic research assistant. Your primary"
            " responsibility is to create thorough, academically"
            " rigorous, unbiased, and systematically organized"
            " reports on a given research topic, following the"
            " standards of scholarly work."
        ),
        "Default Agent": (
            "You are an AI critical thinker research assistant. Your"
            " sole purpose is to write well written, critically"
            " acclaimed, objective and structured reports on given"
            " text."
        ),
    }

    return prompts.get(agent, "No such agent")


def generate_report_prompt(question, research_summary):
    """Generates the report prompt for the given question and research summary.
    Args: question (str): The question to generate the report prompt for
            research_summary (str): The research summary to generate the report prompt for
    Returns: str: The report prompt for the given question and research summary
    """

    return (
        f'"""{research_summary}""" Using the above information,'
        f' answer the following question or topic: "{question}" in a'
        " detailed report -- The report should focus on the answer"
        " to the question, should be well structured, informative,"
        " in depth, with facts and numbers if available, a minimum"
        " of 1,200 words and with markdown syntax and apa format."
        " Write all source urls at the end of the report in apa"
        " format"
    )


def generate_search_queries_prompt(question):
    """Generates the search queries prompt for the given question.
    Args: question (str): The question to generate the search queries prompt for
    Returns: str: The search queries prompt for the given question
    """

    return (
        "Write 4 google search queries to search online that form an"
        f' objective opinion from the following: "{question}"You must'
        " respond with a list of strings in the following format:"
        ' ["query 1", "query 2", "query 3", "query 4"]'
    )


def generate_resource_report_prompt(question, research_summary):
    """Generates the resource report prompt for the given question and research summary.

    Args:
        question (str): The question to generate the resource report prompt for.
        research_summary (str): The research summary to generate the resource report prompt for.

    Returns:
        str: The resource report prompt for the given question and research summary.
    """
    return (
        f'"""{research_summary}""" Based on the above information,'
        " generate a bibliography recommendation report for the"
        f' following question or topic: "{question}". The report'
        " should provide a detailed analysis of each recommended"
        " resource, explaining how each source can contribute to"
        " finding answers to the research question. Focus on the"
        " relevance, reliability, and significance of each source."
        " Ensure that the report is well-structured, informative,"
        " in-depth, and follows Markdown syntax. Include relevant"
        " facts, figures, and numbers whenever available. The report"
        " should have a minimum length of 1,200 words."
    )


def generate_outline_report_prompt(question, research_summary):
    """Generates the outline report prompt for the given question and research summary.
    Args: question (str): The question to generate the outline report prompt for
            research_summary (str): The research summary to generate the outline report prompt for
    Returns: str: The outline report prompt for the given question and research summary
    """

    return (
        f'"""{research_summary}""" Using the above information,'
        " generate an outline for a research report in Markdown"
        f' syntax for the following question or topic: "{question}".'
        " The outline should provide a well-structured framework for"
        " the research report, including the main sections,"
        " subsections, and key points to be covered. The research"
        " report should be detailed, informative, in-depth, and a"
        " minimum of 1,200 words. Use appropriate Markdown syntax to"
        " format the outline and ensure readability."
    )


def generate_concepts_prompt(question, research_summary):
    """Generates the concepts prompt for the given question.
    Args: question (str): The question to generate the concepts prompt for
            research_summary (str): The research summary to generate the concepts prompt for
    Returns: str: The concepts prompt for the given question
    """

    return (
        f'"""{research_summary}""" Using the above information,'
        " generate a list of 5 main concepts to learn for a research"
        f' report on the following question or topic: "{question}".'
        " The outline should provide a well-structured frameworkYou"
        " must respond with a list of strings in the following"
        ' format: ["concepts 1", "concepts 2", "concepts 3",'
        ' "concepts 4, concepts 5"]'
    )


def generate_lesson_prompt(concept):
    """
    Generates the lesson prompt for the given question.
    Args:
        concept (str): The concept to generate the lesson prompt for.
    Returns:
        str: The lesson prompt for the given concept.
    """

    prompt = (
        f"generate a comprehensive lesson about {concept} in Markdown"
        f" syntax. This should include the definitionof {concept},"
        " its historical background and development, its"
        " applications or uses in differentfields, and notable"
        f" events or facts related to {concept}."
    )

    return prompt


def get_report_by_type(report_type):
    report_type_mapping = {
        "research_report": generate_report_prompt,
        "resource_report": generate_resource_report_prompt,
        "outline_report": generate_outline_report_prompt,
    }
    return report_type_mapping[report_type]



from typing import Dict, List, Sequence

from swarms.tools.base_tool import BaseTool
from pydantic import BaseModel


class Step(BaseModel):
    """
    Represents a step in a process.

    Attributes:
        task (str): The task associated with the step.
        id (int): The unique identifier of the step.
        dep (List[int]): The list of step IDs that this step depends on.
        args (Dict[str, str]): The arguments associated with the step.
        tool (BaseTool): The tool used to execute the step.
    """

    task: str = None
    id: int = 0
    dep: List[int] = []
    args: Dict[str, str] = {}
    tool: BaseTool = None
    tools: Sequence[BaseTool] = []
    metadata: Dict[str, str] = {}



from swarms.models.base_llm import BaseLLM
from pydantic import BaseModel
from typing import List, Dict
import openai


class OpenRouterRequest(BaseModel):
    model: str
    messages: List[Dict[str, str]] = []


class OpenRouterChat(BaseLLM):
    """
    A class representing an OpenRouter chat model.

    Args:
        model_name (str): The name of the OpenRouter model.
        base_url (str, optional): The base URL for the OpenRouter API. Defaults to "https://openrouter.ai/api/v1/chat/completions".
        openrouter_api_key (str, optional): The API key for accessing the OpenRouter API. Defaults to None.
        system_prompt (str, optional): The system prompt for the chat model. Defaults to None.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Attributes:
        model_name (str): The name of the OpenRouter model.
        base_url (str): The base URL for the OpenRouter API.
        openrouter_api_key (str): The API key for accessing the OpenRouter API.
        system_prompt (str): The system prompt for the chat model.

    Methods:
        run(task, *args, **kwargs): Runs the chat model with the given task.

    """

    def __init__(
        self,
        model_name: str,
        base_url: str = "https://openrouter.ai/api/v1/chat/completions",
        openrouter_api_key: str = None,
        system_prompt: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.model_name = model_name
        self.base_url = base_url
        self.openrouter_api_key = openrouter_api_key
        self.system_prompt = system_prompt

        openai.api_base = "https://openrouter.ai/api/v1"
        openai.api_key = openrouter_api_key

    def run(self, task: str, *args, **kwargs) -> str:
        """
        Runs the chat model with the given task.

        Args:
            task (str): The user's task for the chat model.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            str: The response generated by the chat model.

        """
        response = openai.ChatCompletion.create(
            model=self.model_name,
            messages=[
                {"role": "system", "content": self.system_prompt},
                {"role": "user", "content": task},
            ]
            * args,
            **kwargs,
        )
        return response.choices[0].message.text



from typing import Union, Sequence, List, Callable
from swarms.structs.agent import Agent
from swarms.structs.base_swarm import BaseSwarm


class SermonSwarm(BaseSwarm):
    """
    Represents a swarm of agents that communicate through sermons.

    Args:
        priest (Agent): The priest agent responsible for generating sermons.
        agents (Sequence[Agent]): The list of agents in the swarm.
        max_loops (int, optional): The maximum number of loops to run the agents. Defaults to 5.
        stop_condition (Union[str, List[str]], optional): The condition(s) that can stop the agents.
            Defaults to "stop".
        stop_function (Union[None, Callable], optional): The function to apply to the sermons before
            checking the stop condition. Defaults to None.
    """

    def __init__(
        self,
        priest: Agent,
        agents: Sequence[Agent],
        max_loops: int = 5,
        stop_condition: Union[str, List[str]] = "stop",
        stop_function: Union[None, Callable] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.priest = priest
        self.agents = agents
        self.max_loops = max_loops
        self.stop_condition = stop_condition
        self.stop_function = stop_function

    def run(self, task: str, *args, **kwargs):
        """
        Runs the swarm by generating sermons from the priest and executing the task on each agent.

        Args:
            task (str): The task to be executed by the agents.
            *args: Additional positional arguments for the task.
            **kwargs: Additional keyword arguments for the task.
        """
        sermon = self.priest(task, *args, **kwargs)

        # Add the sermon to the memory of all agents
        for agent in self.agents:
            agent.add_message_to_memory(sermon)

        # Then run the agents
        for _ in range(self.max_loops):
            for agent in self.agents:
                preach = agent.run(task, *args, **kwargs)

                if self.stop_function:
                    preach = self.stop_function(preach)

                if self.stop_condition in preach:
                    if self.stop_condition is True:
                        break

                    elif self.stop_condition in preach:
                        break



import asyncio
import concurrent.futures
import json
import os
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from typing import Any, Dict, List, Optional

import psutil

try:
    import gzip
except ImportError as error:
    print(f"Error importing gzip: {error}")
# from pydantic import BaseModel


class BaseStructure:
    """Base structure.


    Attributes:
        name (Optional[str]): _description_
        description (Optional[str]): _description_
        save_metadata (bool): _description_
        save_artifact_path (Optional[str]): _description_
        save_metadata_path (Optional[str]): _description_
        save_error_path (Optional[str]): _description_

    Methods:
        run: _description_
        save_to_file: _description_
        load_from_file: _description_
        save_metadata: _description_
        load_metadata: _description_
        log_error: _description_
        save_artifact: _description_
        load_artifact: _description_
        log_event: _description_
        run_async: _description_
        save_metadata_async: _description_
        load_metadata_async: _description_
        log_error_async: _description_
        save_artifact_async: _description_
        load_artifact_async: _description_
        log_event_async: _description_
        asave_to_file: _description_
        aload_from_file: _description_
        run_in_thread: _description_
        save_metadata_in_thread: _description_
        run_concurrent: _description_
        compress_data: _description_
        decompres_data: _description_
        run_batched: _description_
        load_config: _description_
        backup_data: _description_
        monitor_resources: _description_
        run_with_resources: _description_
        run_with_resources_batched: _description_

    Examples:
    >>> base_structure = BaseStructure()
    >>> base_structure
    BaseStructure(name=None, description=None, save_metadata=True, save_artifact_path='./artifacts', save_metadata_path='./metadata', save_error_path='./errors')
    """

    def __init__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        save_metadata: bool = True,
        save_artifact_path: Optional[str] = "./artifacts",
        save_metadata_path: Optional[str] = "./metadata",
        save_error_path: Optional[str] = "./errors",
    ):
        super().__init__()
        self.name = name
        self.description = description
        self.save_metadata = save_metadata
        self.save_artifact_path = save_artifact_path
        self.save_metadata_path = save_metadata_path
        self.save_error_path = save_error_path

    def run(self, *args, **kwargs):
        """Run the structure."""

    def save_to_file(self, data: Any, file_path: str):
        """Save data to file.

        Args:
            data (Any): _description_
            file_path (str): _description_
        """
        with open(file_path, "w") as file:
            json.dump(data, file)

    def load_from_file(self, file_path: str) -> Any:
        """Load data from file.

        Args:
            file_path (str): _description_

        Returns:
            Any: _description_
        """
        with open(file_path) as file:
            return json.load(file)

    def save_metadata(self, metadata: Dict[str, Any]):
        """Save metadata to file.

        Args:
            metadata (Dict[str, Any]): _description_
        """
        if self.save_metadata:
            file_path = os.path.join(
                self.save_metadata_path, f"{self.name}_metadata.json"
            )
            self.save_to_file(metadata, file_path)

    def load_metadata(self) -> Dict[str, Any]:
        """Load metadata from file.

        Returns:
            Dict[str, Any]: _description_
        """
        file_path = os.path.join(
            self.save_metadata_path, f"{self.name}_metadata.json"
        )
        return self.load_from_file(file_path)

    def log_error(self, error_message: str):
        """Log error to file.

        Args:
            error_message (str): _description_
        """
        file_path = os.path.join(
            self.save_error_path, f"{self.name}_errors.log"
        )
        with open(file_path, "a") as file:
            file.write(f"{error_message}\n")

    def save_artifact(self, artifact: Any, artifact_name: str):
        """Save artifact to file.

        Args:
            artifact (Any): _description_
            artifact_name (str): _description_
        """
        file_path = os.path.join(
            self.save_artifact_path, f"{artifact_name}.json"
        )
        self.save_to_file(artifact, file_path)

    def load_artifact(self, artifact_name: str) -> Any:
        """Load artifact from file.

        Args:
            artifact_name (str): _description_

        Returns:
            Any: _description_
        """
        file_path = os.path.join(
            self.save_artifact_path, f"{artifact_name}.json"
        )
        return self.load_from_file(file_path)

    def _current_timestamp(self):
        """Current timestamp.

        Returns:
            _type_: _description_
        """
        return datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    def log_event(
        self,
        event: str,
        event_type: str = "INFO",
    ):
        """Log event to file.

        Args:
            event (str): _description_
            event_type (str, optional): _description_. Defaults to "INFO".
        """
        timestamp = self._current_timestamp()
        log_message = f"[{timestamp}] [{event_type}] {event}\n"
        file = os.path.join(
            self.save_metadata_path, f"{self.name}_events.log"
        )
        with open(file, "a") as file:
            file.write(log_message)

    async def run_async(self, *args, **kwargs):
        """Run the structure asynchronously."""
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, self.run, *args, **kwargs)

    async def save_metadata_async(self, metadata: Dict[str, Any]):
        """Save metadata to file asynchronously.

        Args:
            metadata (Dict[str, Any]): _description_
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(
            None, self.save_metadata, metadata
        )

    async def load_metadata_async(self) -> Dict[str, Any]:
        """Load metadata from file asynchronously.

        Returns:
            Dict[str, Any]: _description_
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, self.load_metadata)

    async def log_error_async(self, error_message: str):
        """Log error to file asynchronously.

        Args:
            error_message (str): _description_
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(
            None, self.log_error, error_message
        )

    async def save_artifact_async(self, artifact: Any, artifact_name: str):
        """Save artifact to file asynchronously.

        Args:
            artifact (Any): _description_
            artifact_name (str): _description_
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(
            None, self.save_artifact, artifact, artifact_name
        )

    async def load_artifact_async(self, artifact_name: str) -> Any:
        """Load artifact from file asynchronously.

        Args:
            artifact_name (str): _description_

        Returns:
            Any: _description_
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(
            None, self.load_artifact, artifact_name
        )

    async def log_event_async(
        self,
        event: str,
        event_type: str = "INFO",
    ):
        """Log event to file asynchronously.

        Args:
            event (str): _description_
            event_type (str, optional): _description_. Defaults to "INFO".
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(
            None, self.log_event, event, event_type
        )

    async def asave_to_file(self, data: Any, file: str, *args, **kwargs):
        """Save data to file asynchronously.

        Args:
            data (Any): _description_
            file (str): _description_
        """
        await asyncio.to_thread(
            self.save_to_file,
            data,
            file,
            *args,
        )

    async def aload_from_file(
        self,
        file: str,
    ) -> Any:
        """Async load data from file.

        Args:
            file (str): _description_

        Returns:
            Any: _description_
        """
        return await asyncio.to_thread(self.load_from_file, file)

    def run_in_thread(self, *args, **kwargs):
        """Run the structure in a thread."""
        with concurrent.futures.ThreadPoolExecutor() as executor:
            return executor.submit(self.run, *args, **kwargs)

    def save_metadata_in_thread(self, metadata: Dict[str, Any]):
        """Save metadata to file in a thread.

        Args:
            metadata (Dict[str, Any]): _description_
        """
        with concurrent.futures.ThreadPoolExecutor() as executor:
            return executor.submit(self.save_metadata, metadata)

    def run_concurrent(self, *args, **kwargs):
        """Run the structure concurrently."""
        return asyncio.run(self.run_async(*args, **kwargs))

    def compress_data(
        self,
        data: Any,
    ) -> bytes:
        """Compress data.

        Args:
            data (Any): _description_

        Returns:
            bytes: _description_
        """
        return gzip.compress(json.dumps(data).encode())

    def decompres_data(self, data: bytes) -> Any:
        """Decompress data.

        Args:
            data (bytes): _description_

        Returns:
            Any: _description_
        """
        return json.loads(gzip.decompress(data).decode())

    def run_batched(
        self,
        batched_data: List[Any],
        batch_size: int = 10,
        *args,
        **kwargs,
    ):
        """Run batched data.

        Args:
            batched_data (List[Any]): _description_
            batch_size (int, optional): _description_. Defaults to 10.

        Returns:
            _type_: _description_
        """
        with ThreadPoolExecutor(max_workers=batch_size) as executor:
            futures = [
                executor.submit(self.run, data) for data in batched_data
            ]
            return [future.result() for future in futures]

    def load_config(
        self, config: str = None, *args, **kwargs
    ) -> Dict[str, Any]:
        """Load config from file.

        Args:
            config (str, optional): _description_. Defaults to None.

        Returns:
            Dict[str, Any]: _description_
        """
        return self.load_from_file(config)

    def backup_data(
        self, data: Any, backup_path: str = None, *args, **kwargs
    ):
        """Backup data to file.

        Args:
            data (Any): _description_
            backup_path (str, optional): _description_. Defaults to None.
        """
        timestamp = self._current_timestamp()
        backup_file_path = f"{backup_path}/{timestamp}.json"
        self.save_to_file(data, backup_file_path)

    def monitor_resources(self):
        """Monitor resource usage."""
        memory = psutil.virtual_memory().percent
        cpu_usage = psutil.cpu_percent(interval=1)
        self.log_event(
            f"Resource usage - Memory: {memory}%, CPU: {cpu_usage}%"
        )

    def run_with_resources(self, *args, **kwargs):
        """Run the structure with resource monitoring."""
        self.monitor_resources()
        return self.run(*args, **kwargs)

    def run_with_resources_batched(
        self,
        batched_data: List[Any],
        batch_size: int = 10,
        *args,
        **kwargs,
    ):
        """Run batched data with resource monitoring.

        Args:
            batched_data (List[Any]): _description_
            batch_size (int, optional): _description_. Defaults to 10.

        Returns:
            _type_: _description_
        """
        self.monitor_resources()
        return self.run_batched(batched_data, batch_size, *args, **kwargs)


# x = BaseStructure()

# print(x)



import hashlib
from time import time_ns
from typing import Callable, List, Optional, Sequence, Union

from swarms.structs.agent import Agent
from swarms.utils.loguru_logger import logger


def _hash(input: str):
    """
    Hashes the input string using SHA256 algorithm.

    Args:
        input (str): The string to be hashed.

    Returns:
        str: The hexadecimal representation of the hash value.
    """
    hex_dig = hashlib.sha256(input.encode("utf-8")).hexdigest()
    return hex_dig


def msg_hash(
    agent: Agent, content: str, turn: int, msg_type: str = "text"
):
    """
    Generate a hash value for a message.

    Args:
        agent (Agent): The agent sending the message.
        content (str): The content of the message.
        turn (int): The turn number of the message.
        msg_type (str, optional): The type of the message. Defaults to "text".

    Returns:
        int: The hash value of the message.
    """
    time = time_ns()
    return _hash(
        f"agent: {agent.agent_name}\ncontent: {content}\ntimestamp:"
        f" {str(time)}\nturn: {turn}\nmsg_type: {msg_type}"
    )


class MessagePool:
    """
    A class representing a message pool for agents in a swarm.

    Attributes:
        agents (Optional[Sequence[Agent]]): The list of agents in the swarm.
        moderator (Optional[Agent]): The moderator agent.
        turns (Optional[int]): The number of turns.
        routing_function (Optional[Callable]): The routing function for message distribution.
        show_names (Optional[bool]): Flag indicating whether to show agent names.
        messages (List[Dict]): The list of messages in the pool.

    Examples:
    >>> from swarms.structs.agent import Agent
    >>> from swarms.structs.message_pool import MessagePool
    >>> agent1 = Agent(agent_name="agent1")
    >>> agent2 = Agent(agent_name="agent2")
    >>> agent3 = Agent(agent_name="agent3")
    >>> moderator = Agent(agent_name="moderator")
    >>> agents = [agent1, agent2, agent3]
    >>> message_pool = MessagePool(agents=agents, moderator=moderator, turns=5)
    >>> message_pool.add(agent=agent1, content="Hello, agent2!", turn=1)
    >>> message_pool.add(agent=agent2, content="Hello, agent1!", turn=1)
    >>> message_pool.add(agent=agent3, content="Hello, agent1!", turn=1)
    >>> message_pool.get_all_messages()
    [{'agent': Agent(agent_name='agent1'), 'content': 'Hello, agent2!', 'turn': 1, 'visible_to': 'all', 'logged': True}, {'agent': Agent(agent_name='agent2'), 'content': 'Hello, agent1!', 'turn': 1, 'visible_to': 'all', 'logged': True}, {'agent': Agent(agent_name='agent3'), 'content': 'Hello, agent1!', 'turn': 1, 'visible_to': 'all', 'logged': True}]
    >>> message_pool.get_visible_messages(agent=agent1, turn=1)
    [{'agent': Agent(agent_name='agent1'), 'content': 'Hello, agent2!', 'turn': 1, 'visible_to': 'all', 'logged': True}, {'agent': Agent(agent_name='agent2'), 'content': 'Hello, agent1!', 'turn': 1, 'visible_to': 'all', 'logged': True}, {'agent': Agent(agent_name='agent3'), 'content': 'Hello, agent1!', 'turn': 1, 'visible_to': 'all', 'logged': True}]
    >>> message_pool.get_visible_messages(agent=agent2, turn=1)
    [{'agent': Agent(agent_name='agent1'), 'content': 'Hello, agent2!', 'turn': 1, 'visible_to': 'all', 'logged': True}, {'agent': Agent(agent_name='agent2'), 'content': 'Hello, agent1!', 'turn': 1, 'visible_to': 'all', 'logged': True}, {'agent': Agent(agent_name='agent3'), 'content': 'Hello, agent1!', 'turn': 1, 'visible_to': 'all', 'logged': True}]
    """

    def __init__(
        self,
        agents: Optional[Sequence[Agent]] = None,
        moderator: Optional[Agent] = None,
        turns: Optional[int] = 5,
        routing_function: Optional[Callable] = None,
        show_names: Optional[bool] = False,
        autosave: Optional[bool] = False,
        *args,
        **kwargs,
    ):
        super().__init__()

        self.agent = agents
        self.moderator = moderator
        self.turns = turns
        self.routing_function = routing_function
        self.show_names = show_names
        self.autosave = autosave

        self.messages = []

        logger.info("MessagePool initialized")
        logger.info(f"Number of agents: {len(agents)}")
        logger.info(f"Agents: {[agent.agent_name for agent in agents]}")
        logger.info(f"moderator: {moderator.agent_name} is available")
        logger.info(f"Number of turns: {turns}")

    def add(
        self,
        agent: Agent,
        content: str,
        turn: int,
        visible_to: Union[str, List[str]] = "all",
        logged: bool = True,
    ):
        """
        Add a message to the pool.

        Args:
            agent (Agent): The agent sending the message.
            content (str): The content of the message.
            turn (int): The turn number.
            visible_to (Union[str, List[str]], optional): The agents who can see the message. Defaults to "all".
            logged (bool, optional): Flag indicating whether the message should be logged. Defaults to True.
        """

        self.messages.append(
            {
                "agent": agent,
                "content": content,
                "turn": turn,
                "visible_to": visible_to,
                "logged": logged,
            }
        )
        logger.info(f"Message added: {content}")

    def reset(self):
        """
        Reset the message pool.
        """
        self.messages = []
        logger.info("MessagePool reset")

    def last_turn(self):
        """
        Get the last turn number.

        Returns:
            int: The last turn number.
        """
        if len(self.messages) == 0:
            return 0
        else:
            return self.messages[-1]["turn"]

    @property
    def last_message(self):
        """
        Get the last message in the pool.

        Returns:
            dict: The last message.
        """
        if len(self.messages) == 0:
            return None
        else:
            return self.messages[-1]

    def get_all_messages(self):
        """
        Get all messages in the pool.

        Returns:
            List[Dict]: The list of all messages.
        """
        return self.messages

    def get_visible_messages(self, agent: Agent, turn: int):
        """
        Get the visible messages for a given agent and turn.

        Args:
            agent (Agent): The agent.
            turn (int): The turn number.

        Returns:
            List[Dict]: The list of visible messages.
        """
        # Get the messages before the current turn
        prev_messages = [
            message for message in self.messages if message["turn"] < turn
        ]

        visible_messages = []
        for message in prev_messages:
            if (
                message["visible_to"] == "all"
                or agent.agent_name in message["visible_to"]
            ):
                visible_messages.append(message)
        return visible_messages

    # def query(self, query: str):
    #     """
    #     Query a message from the messages list and then pass it to the moderator
    #     """
    #     return [
    #         (mod, content)
    #         for mod, content, _ in self.messages  # Add an underscore to ignore the rest of the elements
    #         if query in content
    #     ]



from typing import Any, Optional, Callable

from swarms.structs.agent import Agent
from swarms.tools.json_former import Jsonformer
from swarms.utils.loguru_logger import logger


class ToolAgent(Agent):
    """
    Represents a tool agent that performs a specific task using a model and tokenizer.

    Args:
        name (str): The name of the tool agent.
        description (str): A description of the tool agent.
        model (Any): The model used by the tool agent.
        tokenizer (Any): The tokenizer used by the tool agent.
        json_schema (Any): The JSON schema used by the tool agent.
        *args: Variable length arguments.
        **kwargs: Keyword arguments.

    Attributes:
        name (str): The name of the tool agent.
        description (str): A description of the tool agent.
        model (Any): The model used by the tool agent.
        tokenizer (Any): The tokenizer used by the tool agent.
        json_schema (Any): The JSON schema used by the tool agent.

    Methods:
        run: Runs the tool agent for a specific task.

    Raises:
        Exception: If an error occurs while running the tool agent.


    Example:
        from transformers import AutoModelForCausalLM, AutoTokenizer
        from swarms import ToolAgent


        model = AutoModelForCausalLM.from_pretrained("databricks/dolly-v2-12b")
        tokenizer = AutoTokenizer.from_pretrained("databricks/dolly-v2-12b")

        json_schema = {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "age": {"type": "number"},
                "is_student": {"type": "boolean"},
                "courses": {
                    "type": "array",
                    "items": {"type": "string"}
                }
            }
        }

        task = "Generate a person's information based on the following schema:"
        agent = ToolAgent(model=model, tokenizer=tokenizer, json_schema=json_schema)
        generated_data = agent.run(task)

        print(generated_data)
    """

    def __init__(
        self,
        name: str = "Function Calling Agent",
        description: str = "Generates a function based on the input json schema and the task",
        model: Any = None,
        tokenizer: Any = None,
        json_schema: Any = None,
        max_number_tokens: int = 500,
        parsing_function: Optional[Callable] = None,
        llm: Any = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            agent_name=name,
            agent_description=description,
            llm=llm,
            **kwargs,
        )
        self.name = name
        self.description = description
        self.model = model
        self.tokenizer = tokenizer
        self.json_schema = json_schema
        self.max_number_tokens = max_number_tokens
        self.parsing_function = parsing_function

    def run(self, task: str, *args, **kwargs):
        """
        Run the tool agent for the specified task.

        Args:
            task (str): The task to be performed by the tool agent.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            The output of the tool agent.

        Raises:
            Exception: If an error occurs during the execution of the tool agent.
        """
        try:
            if self.model:
                logger.info(f"Running {self.name} for task: {task}")
                self.toolagent = Jsonformer(
                    model=self.model,
                    tokenizer=self.tokenizer,
                    json_schema=self.json_schema,
                    llm=self.llm,
                    prompt=task,
                    max_number_tokens=self.max_number_tokens,
                    *args,
                    **kwargs,
                )

                if self.parsing_function:
                    out = self.parsing_function(self.toolagent())
                else:
                    out = self.toolagent()

                return out
            elif self.llm:
                logger.info(f"Running {self.name} for task: {task}")
                self.toolagent = Jsonformer(
                    json_schema=self.json_schema,
                    llm=self.llm,
                    prompt=task,
                    max_number_tokens=self.max_number_tokens,
                    *args,
                    **kwargs,
                )

                if self.parsing_function:
                    out = self.parsing_function(self.toolagent())
                else:
                    out = self.toolagent()

                return out

            else:
                raise Exception(
                    "Either model or llm should be provided to the"
                    " ToolAgent"
                )

        except Exception as error:
            logger.error(f"Error running {self.name} for task: {task}")
            raise error



from functools import wraps
import concurrent.futures
from typing import Callable, Any, Dict, List
from swarms.tools.py_func_to_openai_func_str import (
    get_openai_function_schema_from_func,
)
from swarms.utils.loguru_logger import logger


def tool(
    name: str = None,
    description: str = None,
    return_dict: bool = True,
    verbose: bool = True,
    return_string: bool = False,
    return_yaml: bool = False,
):
    """
    A decorator function that generates an OpenAI function schema.

    Args:
        name (str, optional): The name of the OpenAI function. Defaults to None.
        description (str, optional): The description of the OpenAI function. Defaults to None.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        dict: The generated OpenAI function schema.

    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                # Log the function call
                logger.info(f"Creating Tool: {func.__name__}")

                # Assert that the arguments are of the correct type
                assert isinstance(name, str), "name must be a string"
                assert isinstance(
                    description, str
                ), "description must be a string"
                assert isinstance(
                    return_dict, bool
                ), "return_dict must be a boolean"
                assert isinstance(
                    verbose, bool
                ), "verbose must be a boolean"

                # Call the function
                func(*args, **kwargs)

                # Get the openai function schema
                schema = get_openai_function_schema_from_func(
                    func, name=name, description=description
                )

                # Return the schema
                if return_dict:
                    return schema
                elif return_string is True:
                    return str(schema)
                elif return_yaml is True:
                    # schema = YamlModel().dict_to_yaml(schema)
                    return schema
                else:
                    return schema

            except AssertionError as e:
                # Log the assertion error
                logger.error(f"Assertion error: {str(e)}")
                raise

            except Exception as e:
                # Log the exception
                logger.error(f"Exception occurred: {str(e)}")
                raise

        return wrapper

    return decorator


def openai_tool_executor(
    tools: List[Dict[str, Any]],
    function_map: Dict[str, Callable],
    verbose: bool = True,
    *args,
    **kwargs,
) -> Callable:
    """
    Creates a function that dynamically and concurrently executes multiple functions based on parameters specified
    in a list of tool dictionaries, with extensive error handling and validation.

    Args:
        tools (List[Dict[str, Any]]): A list of dictionaries, each containing configuration for a tool, including parameters.
        function_map (Dict[str, Callable]): A dictionary mapping function names to their corresponding callable functions.

    Returns:
        Callable: A function that, when called, executes the specified functions concurrently with the parameters given.

    Examples:
    >>> from swarms.tools.openai_tool_creator_decorator import openai_tool_executor
    >>> from swarms.tools.py_func_to_openai_func_str import get_openai_function_schema_from_func
    >>> from swarms.utils.loguru_logger import logger
    >>>
    >>> def test_function(param1: int, param2: str) -> str:
    ...     return f"Test function called with parameters: {param1}, {param2}"
    ...
    >>> @openai_tool_executor(
    ...     tools=[
    ...         {
    ...             "type": "function",
    ...             "function": {
    ...                 "name": "test_function",
    ...                 "parameters": {
    ...                     "properties": {
    ...                         "param1": {
    ...                             "type": "int",
    ...                             "description": "An integer parameter."
    ...                         },
    ...                         "param2": {
    ...                             "type": "str",
    ...                             "description": "A string parameter."
    ...                         },
    ...                     }
    ...                 }
    ...             }
    ...         }
    ...     ],
    ...     function_map={
    ...         "test_function": test_function
    ...     }
    ... )
    ... def tool_executor():
    ...     pass
    ...
    >>> results = tool_executor()
    >>> logger.info(results)

    """

    def tool_executor():
        # Prepare tasks for concurrent execution
        results = []
        logger.info(f"Executing {len(tools)} tools concurrently.")
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for tool in tools:
                if tool.get("type") != "function":
                    continue  # Skip non-function tool entries

                function_info = tool.get("function", {})
                func_name = function_info.get("name")
                logger.info(f"Executing function: {func_name}")

                # Check if the function name is mapped to an actual function
                if func_name not in function_map:
                    raise KeyError(
                        f"Function '{func_name}' not found in function map."
                    )

                # Validate parameters
                params = function_info.get("parameters", {}).get(
                    "properties", {}
                )
                if not params:
                    raise ValueError(
                        f"No parameters specified for function '{func_name}'."
                    )

                # Submit the function for execution
                try:
                    future = executor.submit(
                        function_map[func_name], **params
                    )
                    futures.append(future)
                except Exception as e:
                    print(
                        f"Failed to submit the function '{func_name}' for execution: {e}"
                    )

            # Gather results from all futures
            for future in futures:
                try:
                    result = future.result()  # Collect result from future
                    results.append(result)
                except Exception as e:
                    print(f"Error during execution of a function: {e}")

        logger.info(f"Results: {results}")

        return results

    return tool_executor



import os
import subprocess as sp
from pathlib import Path

from dotenv import load_dotenv
from PIL import Image

from swarms.models.base_multimodal_model import BaseMultiModalModel

try:
    import google.generativeai as genai
    from google.generativeai.types import GenerationConfig
except ImportError as error:
    print(f"Error importing google.generativeai: {error}")
    print("Please install the google.generativeai package")
    print("pip install google-generativeai")
    sp.run(["pip", "install", "--upgrade", "google-generativeai"])


load_dotenv()


# Helpers
def get_gemini_api_key_env():
    """Get the Gemini API key from the environment

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """
    key = os.getenv("GEMINI_API_KEY")
    if key is None:
        raise ValueError("Please provide a Gemini API key")
    return str(key)


# Main class
class Gemini(BaseMultiModalModel):
    """Gemini model

    Args:
        model_name (str, optional): _description_. Defaults to "gemini-pro".
        gemini_api_key (str, optional): _description_. Defaults to get_gemini_api_key_env.
        return_safety (bool, optional): _description_. Defaults to False.
        candidates (bool, optional): _description_. Defaults to False.
        stream (bool, optional): _description_. Defaults to False.
        candidate_count (int, optional): _description_. Defaults to 1.
        stop_sequence ([type], optional): _description_. Defaults to ['x'].
        max_tokens (int, optional): _description_. Defaults to 100.
        temperature (float, optional): _description_. Defaults to 0.9.

    Methods:
        run: Run the Gemini model
        process_img: Process the image
        chat: Chat with the Gemini model
        list_models: List the Gemini models
        stream_tokens: Stream the tokens
        process_img_pil: Process img



    Examples:
        >>> from swarms.models import Gemini
        >>> gemini = Gemini()
        >>> gemini.run(
                task="A dog",
                img="dog.png",
            )
    """

    def __init__(
        self,
        model_name: str = "gemini-pro-vision",
        gemini_api_key: str = get_gemini_api_key_env,
        return_safety: bool = False,
        candidates: bool = False,
        stream: bool = False,
        candidate_count: int = 1,
        transport: str = "rest",
        stop_sequence=["x"],
        max_tokens: int = 100,
        temperature: float = 0.9,
        system_prompt: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(model_name, *args, **kwargs)
        self.model_name = model_name
        self.gemini_api_key = gemini_api_key
        self.safety = return_safety
        self.candidates = candidates
        self.stream = stream
        self.candidate_count = candidate_count
        self.stop_sequence = stop_sequence
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.system_prompt = system_prompt

        # Configure the API key
        genai.configure(api_key=gemini_api_key, transport=transport)

        # Prepare the generation config
        self.generation_config = GenerationConfig(
            candidate_count=candidate_count,
            # stop_sequence=stop_sequence,
            max_output_tokens=max_tokens,
            temperature=temperature,
            *args,
            **kwargs,
        )

        # Initialize the model
        self.model = genai.GenerativeModel(model_name, *args, **kwargs)

        # Check for the key
        if self.gemini_api_key is None:
            raise ValueError("Please provide a Gemini API key")

    def system_prompt_prep(
        self,
        task: str = None,
        *args,
        **kwargs,
    ):
        """System prompt

        Args:
            system_prompt (str, optional): _description_. Defaults to None.
        """
        PROMPT = f"""
        
        {self.system_prompt}
        
        ######
        
        {task}
        
        """
        return PROMPT

    def run(
        self,
        task: str = None,
        img: str = None,
        *args,
        **kwargs,
    ) -> str:
        """Run the Gemini model

        Args:
            task (str, optional): textual task. Defaults to None.
            img (str, optional): img. Defaults to None.

        Returns:
            str: output from the model
        """
        try:
            prepare_prompt = self.system_prompt_prep(task)
            if img:
                # process_img = self.process_img(img, *args, **kwargs)
                process_img = self.process_img_pil(img)
                response = self.model.generate_content(
                    contents=[prepare_prompt, process_img],
                    generation_config=self.generation_config,
                    stream=self.stream,
                    *args,
                    **kwargs,
                )
                return response.text
            else:
                response = self.model.generate_content(
                    prepare_prompt,
                    stream=self.stream,
                    *args,
                    **kwargs,
                )
                return response.text
        except Exception as error:
            print(f"Error running Gemini model: {error}")
            print(f"Please check the task and image: {task}, {img}")
            raise error

    def process_img(
        self,
        img: str = None,
        type: str = "image/png",
        *args,
        **kwargs,
    ):
        """Process the image

        Args:
            img (str, optional): _description_. Defaults to None.
            type (str, optional): _description_. Defaults to "image/png".

        Raises:
            ValueError: _description_
            ValueError: _description_
            ValueError: _description_
        """
        try:
            if img is None:
                raise ValueError("Please provide an image to process")
            if type is None:
                raise ValueError("Please provide the image type")
            if self.gemini_api_key is None:
                raise ValueError("Please provide a Gemini API key")

            # Load the image
            img = [{"mime_type": type, "data": Path(img).read_bytes()}]
        except Exception as error:
            print(f"Error processing image: {error}")

    def chat(
        self,
        task: str = None,
        img: str = None,
        *args,
        **kwargs,
    ) -> str:
        """Chat with the Gemini model

        Args:
            task (str, optional): _description_. Defaults to None.
            img (str, optional): _description_. Defaults to None.

        Returns:
            str: _description_
        """
        chat = self.model.start_chat()
        response = chat.send_message(task, *args, **kwargs)
        response1 = response.text
        print(response1)
        response = chat.send_message(img, *args, **kwargs)

    def list_models(self) -> str:
        """List the Gemini models

        Returns:
            str: _description_
        """
        for m in genai.list_models():
            if "generateContent" in m.supported_generation_methods:
                print(m.name)

    def stream_tokens(self, content: str = None):
        """Stream the tokens

        Args:
            content (t, optional): _description_. Defaults to None.
        """
        for chunk in content:
            print(chunk.text)
            print("_" * 80)

    def process_img_pil(self, img: str = None):
        """Process img

        Args:
            img (str, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        img = Image.open(img)
        return img



import logging
from typing import List

from swarms.structs.base_structure import BaseStructure
from swarms.structs.task import Task

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class RecursiveWorkflow(BaseStructure):
    """
    RecursiveWorkflow class for running a task recursively until a stopping condition is met.

    Args:
        task (Task): The task to execute.
        stop_token (Any): The token that indicates when to stop the workflow.

    Attributes:
        task (Task): The task to execute.
        stop_token (Any): The token that indicates when to stop the workflow.

    Examples:
    >>> from swarms.models import OpenAIChat
    >>> from swarms.structs import RecursiveWorkflow, Task
    >>> llm = OpenAIChat(openai_api_key="")
    >>> task = Task(llm, "What's the weather in miami")
    >>> workflow = RecursiveWorkflow()
    >>> workflow.add(task)
    >>> workflow.run()
    """

    def __init__(
        self,
        stop_token: str = "<DONE>",
        stopping_conditions: callable = None,
    ):
        self.stop_token = stop_token
        self.stopping_conditions = stopping_conditions
        self.task_pool = []

        assert self.stop_token is not None, "stop_token cannot be None"

    def add(self, task: Task = None, tasks: List[Task] = None):
        """Adds a task to the workflow.

        Args:
            task (Task): _description_
            tasks (List[Task]): _description_
        """
        try:
            if tasks:
                for task in tasks:
                    if isinstance(task, Task):
                        self.task_pool.append(task)
                        logger.info(
                            "[INFO][RecursiveWorkflow] Added task"
                            f" {task} to workflow"
                        )
            elif isinstance(task, Task):
                self.task_pool.append(task)
                logger.info(
                    f"[INFO][RecursiveWorkflow] Added task {task} to"
                    " workflow"
                )
        except Exception as error:
            logger.warning(f"[ERROR][RecursiveWorkflow] {error}")
            raise error

    def run(self):
        """
        Executes the tasks in the workflow until the stop token is encountered.

        Returns:
            None
        """
        try:
            for task in self.task_pool:
                while True:
                    result = task.run()
                    if result is not None and self.stop_token in result:
                        break
                    print(f"{result}")
        except Exception as error:
            logger.warning(f"[ERROR][RecursiveWorkflow] {error}")
            raise error



import inspect


def process_tool_docs(item):
    """
    Process the documentation for a given item.

    Args:
        item: The item to process the documentation for.

    Returns:
        metadata: The processed metadata containing the item's name, documentation, and source code.
    """
    # If item is an instance of a class, get its class
    if not inspect.isclass(item) and hasattr(item, "__class__"):
        item = item.__class__

    doc = inspect.getdoc(item)
    source = inspect.getsource(item)
    is_class = inspect.isclass(item)
    item_type = "Class Name" if is_class else "Function Name"
    metadata = f"{item_type}: {item.__name__}\n\n"
    if doc:
        metadata += f"Documentation:\n{doc}\n\n"
    metadata += f"\n{source}"
    return metadata



from typing import List

from swarms.structs.agent import Agent
from swarms.utils.parse_code import extract_code_from_markdown


class LongContextSwarmLeader:
    """
    Represents a leader in a long context swarm.

    Args:
        - llm (str): The language model to use for the agent.
        - agents (List[Agent]): The agents in the swarm.
        - prompt_template_json (str): The SOP template in JSON format.
        - return_parsed (bool): Whether to return the parsed output.

    """

    def __init__(
        self,
        llm,
        agents: List[Agent] = None,
        prompt_template_json: str = None,
        return_parsed: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.llm = llm
        self.agents = agents
        self.prompt_template_json = prompt_template_json
        self.return_parsed = return_parsed

        # Create an instance of the Agent class
        self.agent = Agent(
            llm=llm,
            system_prompt=None,
            sop=self.prompt_template_json,
            *args,
            **kwargs,
        )

    def prep_schema(self, task: str, *args, **kwargs):
        """
        Returns a formatted string containing the metadata of all agents in the swarm.

        Parameters:
        - task (str): The description of the task.

        Returns:
        - prompt (str): The formatted string containing the agent metadata.
        """
        prompt = f"""
        
        You need to recruit a team of members to solve a
        task. Select the appropriate member based on the
        task description:
        
        # Task Description
        {task}
        
        # Members
        
        Your output must follow this JSON schema below in markdown format:
            {{
                "agent_id": "string",
                "agent_name": "string",
                "agent_description": "string"
            }}
        
        """
        for agent in self.agents:
            prompt += (
                f"Member Name: {agent.ai_name}\nMember ID:"
                f" {agent.id}\nMember Description:"
                f" {agent.description}\n\n"
            )

        return prompt

    def prep_schema_second(self, task_description: str, task: str):
        prompt = f"""
        You are the leader of a team of {len(self.agents)}
        members. Your team will need to collaborate to
        solve a task. The rule is:
        
        1. Only you know the task description and task
        objective; the other members do not.
        2. But they will receive different documents that
        may contain answers, and you need to send them
        an instruction to query their document.
        3. Your instruction need to include your
        understanding of the task and what you need them
        to focus on. If necessary, your instructions can
        explicitly include the task objective.
        4. Finally, you need to complete the task based on
        the query results they return.
        
        # Task Description:
        {task_description}
        
        # Task Objective:
        {task}
        
        # Generate Instruction for Members:
        Now, you need to generate an instruction for all
        team members. You can ask them to answer a
        certain question, or to extract information related
        to the task, based on their respective documents.
        Your output must following the JSON
        format: {{"type": "instruction", "content":
        "your_instruction_content"}}
        
        """
        return prompt

    def run(self, task: str, *args, **kwargs):
        """
        Executes the specified task using the agent's run method.

        Args:
            task: The task to be executed.
            *args: Additional positional arguments for the task.
            **kwargs: Additional keyword arguments for the task.

        Returns:
            The result of the task execution.
        """
        task = self.prep_schema(task)
        out = self.agent.run(task, *args, **kwargs)

        if self.return_parsed:
            out = extract_code_from_markdown(out)

        return out


# class LongContextSwarm(BaseSwarm):
#     def __init__(
#         self,
#         agents: List[Agent],
#         Leader: Agent,
#         team_loops: int,
#         *args,
#         **kwargs,
#     ):
#         super().__init__()
#         self.agents = agents
#         self.leader = Leader
#         self.team_loops = team_loops
#         self.chunks = len(agents)



import os

import cv2
import numpy as np
import requests
import torch
from PIL import Image
from transformers import AutoModelForVision2Seq, AutoProcessor

from swarms.models.base_multimodal_model import BaseMultiModalModel


# utils
def is_overlapping(rect1, rect2):
    x1, y1, x2, y2 = rect1
    x3, y3, x4, y4 = rect2
    return not (x2 < x3 or x1 > x4 or y2 < y3 or y1 > y4)


class Kosmos(BaseMultiModalModel):
    """
    Kosmos model by Yen-Chun Shieh

    Parameters
    ----------
    model_name : str
        Path to the pretrained model

    Examples
    --------
    >>> kosmos = Kosmos()
    >>> kosmos("Hello, my name is", "path/to/image.png")

    """

    def __init__(
        self,
        model_name="ydshieh/kosmos-2-patch14-224",
        max_new_tokens: int = 64,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.max_new_tokens = max_new_tokens

        self.model = AutoModelForVision2Seq.from_pretrained(
            model_name, trust_remote_code=True, *args, **kwargs
        )
        self.processor = AutoProcessor.from_pretrained(
            model_name, trust_remote_code=True, *args, **kwargs
        )

    def get_image(self, url: str):
        """Get image from url

        Args:
            url (str): url of image

        Returns:
            _type_: _description_
        """
        return Image.open(requests.get(url, stream=True).raw)

    def run(self, task: str, image: str, *args, **kwargs):
        """Run the model

        Args:
            task (str): task to run
            image (str): img url
        """
        inputs = self.processor(
            text=task, images=image, return_tensors="pt"
        )
        generated_ids = self.model.generate(
            pixel_values=inputs["pixel_values"],
            input_ids=inputs["input_ids"][:, :-1],
            attention_mask=inputs["attention_mask"][:, :-1],
            image_embeds=None,
            img_attn_mask=inputs["img_attn_mask"][:, :-1],
            use_cache=True,
            max_new_tokens=self.max_new_tokens,
        )

        generated_texts = self.processor.batch_decode(
            generated_ids,
            skip_special_tokens=True,
        )[0]

        processed_text, entities = self.processor.post_process_generation(
            generated_texts
        )

        return processed_text, entities

    # tasks
    def multimodal_grounding(self, phrase, image_url):
        task = f"<grounding><phrase> {phrase} </phrase>"
        self.run(task, image_url)

    def referring_expression_comprehension(self, phrase, image_url):
        task = f"<grounding><phrase> {phrase} </phrase>"
        self.run(task, image_url)

    def referring_expression_generation(self, phrase, image_url):
        task = (
            "<grounding><phrase>"
            " It</phrase><object><patch_index_0044><patch_index_0863></object> is"
        )
        self.run(task, image_url)

    def grounded_vqa(self, question, image_url):
        task = f"<grounding> Question: {question} Answer:"
        self.run(task, image_url)

    def grounded_image_captioning(self, image_url):
        task = "<grounding> An image of"
        self.run(task, image_url)

    def grounded_image_captioning_detailed(self, image_url):
        task = "<grounding> Describe this image in detail"
        self.run(task, image_url)

    def draw_entity_boxes_on_image(
        image, entities, show=False, save_path=None
    ):
        """_summary_
        Args:
            image (_type_): image or image path
            collect_entity_location (_type_): _description_
        """
        if isinstance(image, Image.Image):
            image_h = image.height
            image_w = image.width
            image = np.array(image)[:, :, [2, 1, 0]]
        elif isinstance(image, str):
            if os.path.exists(image):
                pil_img = Image.open(image).convert("RGB")
                image = np.array(pil_img)[:, :, [2, 1, 0]]
                image_h = pil_img.height
                image_w = pil_img.width
            else:
                raise ValueError(f"invaild image path, {image}")
        elif isinstance(image, torch.Tensor):
            # pdb.set_trace()
            image_tensor = image.cpu()
            reverse_norm_mean = torch.tensor(
                [0.48145466, 0.4578275, 0.40821073]
            )[:, None, None]
            reverse_norm_std = torch.tensor(
                [0.26862954, 0.26130258, 0.27577711]
            )[:, None, None]
            image_tensor = (
                image_tensor * reverse_norm_std + reverse_norm_mean
            )
            # pil_img = T.ToPILImage()(image_tensor)
            image_h = pil_img.height
            image_w = pil_img.width
            image = np.array(pil_img)[:, :, [2, 1, 0]]
        else:
            raise ValueError(
                f"invaild image format, {type(image)} for {image}"
            )

        if len(entities) == 0:
            return image

        new_image = image.copy()
        previous_bboxes = []
        # size of text
        text_size = 1
        # thickness of text
        text_line = 1  # int(max(1 * min(image_h, image_w) / 512, 1))
        box_line = 3
        (c_width, text_height), _ = cv2.getTextSize(
            "F", cv2.FONT_HERSHEY_COMPLEX, text_size, text_line
        )
        base_height = int(text_height * 0.675)
        text_offset_original = text_height - base_height
        text_spaces = 3

        for entity_name, (start, end), bboxes in entities:
            for x1_norm, y1_norm, x2_norm, y2_norm in bboxes:
                orig_x1, orig_y1, orig_x2, orig_y2 = (
                    int(x1_norm * image_w),
                    int(y1_norm * image_h),
                    int(x2_norm * image_w),
                    int(y2_norm * image_h),
                )
                # draw bbox
                # random color
                color = tuple(np.random.randint(0, 255, size=3).tolist())
                new_image = cv2.rectangle(
                    new_image,
                    (orig_x1, orig_y1),
                    (orig_x2, orig_y2),
                    color,
                    box_line,
                )

                l_o, r_o = (
                    box_line // 2 + box_line % 2,
                    box_line // 2 + box_line % 2 + 1,
                )

                x1 = orig_x1 - l_o
                y1 = orig_y1 - l_o

                if (
                    y1
                    < text_height + text_offset_original + 2 * text_spaces
                ):
                    y1 = (
                        orig_y1
                        + r_o
                        + text_height
                        + text_offset_original
                        + 2 * text_spaces
                    )
                    x1 = orig_x1 + r_o

                # add text background
                (text_width, text_height), _ = cv2.getTextSize(
                    f"  {entity_name}",
                    cv2.FONT_HERSHEY_COMPLEX,
                    text_size,
                    text_line,
                )
                text_bg_x1, text_bg_y1, text_bg_x2, text_bg_y2 = (
                    x1,
                    y1
                    - (
                        text_height
                        + text_offset_original
                        + 2 * text_spaces
                    ),
                    x1 + text_width,
                    y1,
                )

                for prev_bbox in previous_bboxes:
                    while is_overlapping(
                        (
                            text_bg_x1,
                            text_bg_y1,
                            text_bg_x2,
                            text_bg_y2,
                        ),
                        prev_bbox,
                    ):
                        text_bg_y1 += (
                            text_height
                            + text_offset_original
                            + 2 * text_spaces
                        )
                        text_bg_y2 += (
                            text_height
                            + text_offset_original
                            + 2 * text_spaces
                        )
                        y1 += (
                            text_height
                            + text_offset_original
                            + 2 * text_spaces
                        )

                        if text_bg_y2 >= image_h:
                            text_bg_y1 = max(
                                0,
                                image_h
                                - (
                                    text_height
                                    + text_offset_original
                                    + 2 * text_spaces
                                ),
                            )
                            text_bg_y2 = image_h
                            y1 = image_h
                            break

                alpha = 0.5
                for i in range(text_bg_y1, text_bg_y2):
                    for j in range(text_bg_x1, text_bg_x2):
                        if i < image_h and j < image_w:
                            if j < text_bg_x1 + 1.35 * c_width:
                                # original color
                                bg_color = color
                            else:
                                # white
                                bg_color = [255, 255, 255]
                            new_image[i, j] = (
                                alpha * new_image[i, j]
                                + (1 - alpha) * np.array(bg_color)
                            ).astype(np.uint8)

                cv2.putText(
                    new_image,
                    f"  {entity_name}",
                    (x1, y1 - text_offset_original - 1 * text_spaces),
                    cv2.FONT_HERSHEY_COMPLEX,
                    text_size,
                    (0, 0, 0),
                    text_line,
                    cv2.LINE_AA,
                )
                # previous_locations.append((x1, y1))
                previous_bboxes.append(
                    (text_bg_x1, text_bg_y1, text_bg_x2, text_bg_y2)
                )

        pil_image = Image.fromarray(new_image[:, :, [2, 1, 0]])
        if save_path:
            pil_image.save(save_path)
        if show:
            pil_image.show()

        return new_image

    def generate_boxees(self, task, image_url):
        image = self.get_image(image_url)
        processed_text, entities = self.process_task(task, image)
        self.draw_entity_boxes_on_image(image, entities, show=True)



from typing import (
    Any,
    Callable,
    Sequence,
    Union,
)
from swarms.models.base_llm import BaseLLM
from swarms.models.base_multimodal_model import BaseMultiModalModel
from swarms.structs.agent import Agent

# Unified type for agent
AgentType = Union[Agent, Callable, Any, BaseLLM, BaseMultiModalModel]

# List of agents
AgentListType = Sequence[AgentType]



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

from swarms.structs.agent import Agent
from swarms.structs.conversation import Conversation
from swarms.utils.logger import logger


@dataclass
class Company:
    """
    Represents a company with a hierarchical organizational structure.
    """

    org_chart: List[List[Agent]]
    shared_instructions: str = None
    ceo: Optional[Agent] = None
    agents: List[Agent] = field(default_factory=list)
    agent_interactions: Dict[str, List[str]] = field(default_factory=dict)
    history: Conversation = field(default_factory=Conversation)

    def __post_init__(self):
        self._parse_org_chart(self.org_chart)

    def add(self, agent: Agent) -> None:
        """
        Adds an agent to the company.

        Args:
            agent (Agent): The agent to be added.

        Raises:
            ValueError: If an agent with the same ID already exists in the company.
        """
        try:
            if any(
                existing_agent.id == agent.id
                for existing_agent in self.agents
            ):
                raise ValueError(
                    f"Agent with id {agent.id} already exists in the"
                    " company."
                )
            self.agents.append(agent)

        except Exception as error:
            logger.error(f"[ERROR][CLASS: Company][METHOD: add] {error}")
            raise error

    def get(self, agent_name: str) -> Agent:
        """
        Retrieves an agent from the company by name.

        Args:
            agent_name (str): The name of the agent to retrieve.

        Returns:
            Agent: The retrieved agent.

        Raises:
            ValueError: If an agent with the specified name does not exist in the company.
        """
        try:
            for agent in self.agents:
                if agent.name == agent_name:
                    return agent
            raise ValueError(
                f"Agent with name {agent_name} does not exist in the"
                " company."
            )
        except Exception as error:
            logger.error(f"[ERROR][CLASS: Company][METHOD: get] {error}")
            raise error

    def remove(self, agent: Agent) -> None:
        """
        Removes an agent from the company.

        Args:
            agent (Agent): The agent to be removed.
        """
        try:
            self.agents.remove(agent)
        except Exception as error:
            logger.error(
                f"[ERROR][CLASS: Company][METHOD: remove] {error}"
            )
            raise error

    def _parse_org_chart(
        self, org_chart: Union[List[Agent], List[List[Agent]]]
    ) -> None:
        """
        Parses the organization chart and adds agents to the company.

        Args:
            org_chart (Union[List[Agent], List[List[Agent]]]): The organization chart
                representing the hierarchy of agents.

        Raises:
            ValueError: If more than one CEO is found in the org chart or if an invalid
                agent is encountered.
        """
        try:
            for node in org_chart:
                if isinstance(node, Agent):
                    if self.ceo:
                        raise ValueError("1 CEO is only allowed")
                    self.ceo = node
                    self.add(node)

                elif isinstance(node, list):
                    for agent in node:
                        if not isinstance(agent, Agent):
                            raise ValueError("Invalid agent in org chart")
                        self.add(agent)

                    for i, agent in enumerate(node):
                        if i == len(node) - 1:
                            continue

                        for other_agent in node[i + 1]:
                            self.__init_task(agent, other_agent)
        except Exception as error:
            logger.error(
                "[ERROR][CLASS: Company][METHOD: _parse_org_chart]"
                f" {error}"
            )
            raise error

    def _init_interaction(
        self,
        agent1: Agent,
        agent2: Agent,
    ) -> None:
        """
        Initializes the interaction between two agents.

        Args:
            agent1 (Agent): The first agent involved in the interaction.
            agent2 (Agent): The second agent involved in the interaction.

        Returns:
            None
        """
        if agent1.ai_name not in self.agents_interactions:
            self.agents_interactions[agent1.ai_name] = []
        self.agents_interactions[agent1.ai_name].append(agent2.ai_name)

    def run(self):
        """
        Run the company
        """
        for (
            agent_name,
            interaction_agents,
        ) in self.agents_interactions.items():
            agent = self.get(agent_name)
            for interaction_agent in interaction_agents:
                task_description = (
                    f"Task for {agent_name} to interact with"
                    f" {interaction_agent}"
                )
                print(f"{task_description} is being executed")
                agent.run(task_description)



import datetime
from typing import Dict, Optional


class Message:
    """
    Represents a message with timestamp and optional metadata.

    Usage
    --------------
    mes = Message(
        sender = "Kye",
        content = "message"
    )

    print(mes)
    """

    def __init__(
        self,
        sender: str,
        content: str,
        metadata: Optional[Dict[str, str]] = None,
    ):
        self.timestamp: datetime.datetime = datetime.datetime.now()
        self.sender: str = sender
        self.content: str = content
        self.metadata: Dict[str, str] = metadata or {}

    def __repr__(self) -> str:
        """
        __repr__ means...
        """
        return f"{self.timestamp} - {self.sender}: {self.content}"



import asyncio
import os
import time
from functools import wraps
from typing import Union

import torch
from termcolor import colored
from transformers import (
    AutoModelForSpeechSeq2Seq,
    AutoProcessor,
    pipeline,
)


def async_retry(max_retries=3, exceptions=(Exception,), delay=1):
    """
    A decorator for adding retry logic to async functions.
    :param max_retries: Maximum number of retries before giving up.
    :param exceptions: A tuple of exceptions to catch and retry on.
    :param delay: Delay between retries.
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            retries = max_retries
            while retries:
                try:
                    return await func(*args, **kwargs)
                except exceptions as e:
                    retries -= 1
                    if retries <= 0:
                        raise
                    print(
                        f"Retry after exception: {e}, Attempts"
                        f" remaining: {retries}"
                    )
                    await asyncio.sleep(delay)

        return wrapper

    return decorator


class DistilWhisperModel:
    """
    This class encapsulates the Distil-Whisper model for English speech recognition.
    It allows for both synchronous and asynchronous transcription of short and long-form audio.

    Args:
        model_id: The model ID to use. Defaults to "distil-whisper/distil-large-v2".


    Attributes:
        device: The device to use for inference.
        torch_dtype: The torch data type to use for inference.
        model_id: The model ID to use.
        model: The model instance.
        processor: The processor instance.

    Usage:
        model_wrapper = DistilWhisperModel()
        transcription = model_wrapper('path/to/audio.mp3')

        # For async usage
        transcription = asyncio.run(model_wrapper.async_transcribe('path/to/audio.mp3'))
    """

    def __init__(self, model_id="distil-whisper/distil-large-v2"):
        self.device = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.torch_dtype = (
            torch.float16 if torch.cuda.is_available() else torch.float32
        )
        self.model_id = model_id
        self.model = AutoModelForSpeechSeq2Seq.from_pretrained(
            model_id,
            torch_dtype=self.torch_dtype,
            low_cpu_mem_usage=True,
            use_safetensors=True,
        ).to(self.device)
        self.processor = AutoProcessor.from_pretrained(model_id)

    def __call__(self, inputs: Union[str, dict]):
        return self.transcribe(inputs)

    def transcribe(self, inputs: Union[str, dict]):
        """
        Synchronously transcribe the given audio input using the Distil-Whisper model.
        :param inputs: A string representing the file path or a dict with audio data.
        :return: The transcribed text.
        """
        pipe = pipeline(
            "automatic-speech-recognition",
            model=self.model,
            tokenizer=self.processor.tokenizer,
            feature_extractor=self.processor.feature_extractor,
            max_new_tokens=128,
            torch_dtype=self.torch_dtype,
            device=self.device,
        )

        return pipe(inputs)["text"]

    @async_retry()
    async def async_transcribe(self, inputs: Union[str, dict]):
        """
        Asynchronously transcribe the given audio input using the Distil-Whisper model.
        :param inputs: A string representing the file path or a dict with audio data.
        :return: The transcribed text.
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, self.transcribe, inputs)

    def real_time_transcribe(self, audio_file_path, chunk_duration=5):
        """
        Simulates real-time transcription of an audio file, processing and printing results
        in chunks with colored output for readability.

        :param audio_file_path: Path to the audio file to be transcribed.
        :param chunk_duration: Duration in seconds of each audio chunk to be processed.
        """
        if not os.path.isfile(audio_file_path):
            print(colored("The audio file was not found.", "red"))
            return

        # Assuming `chunk_duration` is in seconds and `processor` can handle chunk-wise processing
        try:
            with torch.no_grad():
                # Load the whole audio file, but process and transcribe it in chunks
                audio_input = self.processor.audio_file_to_array(
                    audio_file_path
                )
                sample_rate = audio_input.sampling_rate
                len(audio_input.array) / sample_rate
                chunks = [
                    audio_input.array[i : i + sample_rate * chunk_duration]
                    for i in range(
                        0,
                        len(audio_input.array),
                        sample_rate * chunk_duration,
                    )
                ]

                print(
                    colored("Starting real-time transcription...", "green")
                )

                for i, chunk in enumerate(chunks):
                    # Process the current chunk
                    processed_inputs = self.processor(
                        chunk,
                        sampling_rate=sample_rate,
                        return_tensors="pt",
                        padding=True,
                    )
                    processed_inputs = processed_inputs.input_values.to(
                        self.device
                    )

                    # Generate transcription for the chunk
                    logits = self.model.generate(processed_inputs)
                    transcription = self.processor.batch_decode(
                        logits, skip_special_tokens=True
                    )[0]

                    # Print the chunk's transcription
                    print(
                        colored(f"Chunk {i+1}/{len(chunks)}: ", "yellow")
                        + transcription
                    )

                    # Wait for the chunk's duration to simulate real-time processing
                    time.sleep(chunk_duration)

        except Exception as e:
            print(
                colored(
                    f"An error occurred during transcription: {e}",
                    "red",
                )
            )



from swarms.utils.loguru_logger import logger


def try_except_wrapper(func, verbose: bool = False):
    """
    A decorator that wraps a function with a try-except block.
    It catches any exception that occurs during the execution of the function,
    prints an error message, and returns None.
    It also prints a message indicating the exit of the function.

    Args:
        func (function): The function to be wrapped.

    Returns:
        function: The wrapped function.

    Examples:
    >>> @try_except_wrapper(verbose=True)
    ... def divide(a, b):
    ...     return a / b
    >>> divide(1, 0)
    An error occurred in function divide: division by zero
    Exiting function: divide
    """

    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
            return result
        except Exception as error:
            if verbose:
                logger.error(
                    f"An error occurred in function {func.__name__}:"
                    f" {error}"
                )
            else:
                print(
                    f"An error occurred in function {func.__name__}:"
                    f" {error}"
                )
                return None
        finally:
            print(f"Exiting function: {func.__name__}")

    return wrapper



import logging
import queue
import threading
from concurrent.futures import (
    FIRST_COMPLETED,
    ThreadPoolExecutor,
    wait,
)
from typing import List

from swarms.structs.base_workflow import BaseWorkflow
from swarms.structs.task import Task

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(message)s",
)


class PriorityTask:
    """
    Represents a task with a priority level.

    Attributes:
        task (Task): The task to be executed.
        priority (int): The priority level of the task.
    """

    def __init__(self, task: Task, priority: int = 0):
        self.task = task
        self.priority = priority

    def __lt__(self, other):
        return self.priority < other.priority


class MultiThreadedWorkflow(BaseWorkflow):
    """
    Represents a multi-threaded workflow that executes tasks concurrently using a thread pool.

    Args:
        max_workers (int): The maximum number of worker threads in the thread pool. Default is 5.
        autosave (bool): Flag indicating whether to automatically save task results. Default is True.
        tasks (List[PriorityTask]): List of priority tasks to be executed. Default is an empty list.
        retry_attempts (int): The maximum number of retry attempts for failed tasks. Default is 3.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Attributes:
        max_workers (int): The maximum number of worker threads in the thread pool.
        autosave (bool): Flag indicating whether to automatically save task results.
        retry_attempts (int): The maximum number of retry attempts for failed tasks.
        tasks_queue (PriorityQueue): The queue that holds the priority tasks.
        lock (Lock): The lock used for thread synchronization.

    Methods:
        execute_tasks: Executes the tasks in the thread pool and returns the results.
        _autosave_task_result: Autosaves the result of a task.

    """

    def __init__(
        self,
        max_workers: int = 5,
        autosave: bool = True,
        tasks: List[PriorityTask] = None,
        retry_attempts: int = 3,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.max_workers = max_workers
        self.autosave = autosave
        self.retry_attempts = retry_attempts
        if tasks is None:
            tasks = []
        self.tasks_queue = queue.PriorityQueue()
        for task in tasks:
            self.tasks_queue.put(task)
        self.lock = threading.Lock()

    def run(self):
        """
        Executes the tasks in the thread pool and returns the results.

        Returns:
            List: The list of results from the executed tasks.

        """
        results = []
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            future_to_task = {}
            for _ in range(self.tasks_queue.qsize()):
                priority_task = self.tasks_queue.get_nowait()
                future = executor.submit(priority_task.task.execute)
                future_to_task[future] = (
                    priority_task.task,
                    0,
                )  # (Task, attempt)

            while future_to_task:
                # Wait for the next future to complete
                done, _ = wait(
                    future_to_task.keys(), return_when=FIRST_COMPLETED
                )

                for future in done:
                    task, attempt = future_to_task.pop(future)
                    try:
                        result = future.result()
                        results.append(result)
                        logging.info(
                            f"Task {task} completed successfully with"
                            f" result: {result}"
                        )
                        if self.autosave:
                            self._autosave_task_result(task, result)
                    except Exception as e:
                        logging.error(
                            (
                                f"Attempt {attempt+1} failed for task"
                                f" {task}: {str(e)}"
                            ),
                            exc_info=True,
                        )
                        if attempt + 1 < self.retry_attempts:
                            # Retry the task
                            retry_future = executor.submit(task.execute)
                            future_to_task[retry_future] = (
                                task,
                                attempt + 1,
                            )
                        else:
                            logging.error(
                                f"Task {task} failed after"
                                f" {self.retry_attempts} attempts."
                            )

        return results

    def _autosave_task_result(self, task: Task, result):
        """
        Autosaves the result of a task.

        Args:
            task (Task): The task whose result needs to be autosaved.
            result: The result of the task.

        """
        with self.lock:
            logging.info(f"Autosaving result for task {task}: {result}")
            # Actual autosave logic goes here



from __future__ import annotations

from abc import abstractmethod
from typing import Sequence


class Message:
    """
    The base abstract Message class.
    Messages are the inputs and outputs of ChatModels.
    """

    def __init__(
        self, content: str, role: str, additional_kwargs: dict = None
    ):
        self.content = content
        self.role = role
        self.additional_kwargs = (
            additional_kwargs if additional_kwargs else {}
        )

    @abstractmethod
    def get_type(self) -> str:
        pass


class HumanMessage(Message):
    """
    A Message from a human.
    """

    def __init__(
        self,
        content: str,
        role: str = "Human",
        additional_kwargs: dict = None,
        example: bool = False,
    ):
        super().__init__(content, role, additional_kwargs)
        self.example = example

    def get_type(self) -> str:
        return "human"


class AIMessage(Message):
    """
    A Message from an AI.
    """

    def __init__(
        self,
        content: str,
        role: str = "AI",
        additional_kwargs: dict = None,
        example: bool = False,
    ):
        super().__init__(content, role, additional_kwargs)
        self.example = example

    def get_type(self) -> str:
        return "ai"


class SystemMessage(Message):
    """
    A Message for priming AI behavior, usually passed in as the first of a sequence
    of input messages.
    """

    def __init__(
        self,
        content: str,
        role: str = "System",
        additional_kwargs: dict = None,
    ):
        super().__init__(content, role, additional_kwargs)

    def get_type(self) -> str:
        return "system"


class FunctionMessage(Message):
    """
    A Message for passing the result of executing a function back to a model.
    """

    def __init__(
        self,
        content: str,
        role: str = "Function",
        name: str = None,
        additional_kwargs: dict = None,
    ):
        super().__init__(content, role, additional_kwargs)
        self.name = name

    def get_type(self) -> str:
        return "function"


class ChatMessage(Message):
    """
    A Message that can be assigned an arbitrary speaker (i.e. role).
    """

    def __init__(
        self, content: str, role: str, additional_kwargs: dict = None
    ):
        super().__init__(content, role, additional_kwargs)

    def get_type(self) -> str:
        return "chat"


def get_buffer_string(
    messages: Sequence[Message],
    human_prefix: str = "Human",
    ai_prefix: str = "AI",
) -> str:
    string_messages = []
    for m in messages:
        message = f"{m.role}: {m.content}"
        if (
            isinstance(m, AIMessage)
            and "function_call" in m.additional_kwargs
        ):
            message += f"{m.additional_kwargs['function_call']}"
        string_messages.append(message)

    return "\n".join(string_messages)


def message_to_dict(message: Message) -> dict:
    return {"type": message.get_type(), "data": message.__dict__}


def messages_to_dict(messages: Sequence[Message]) -> list[dict]:
    return [message_to_dict(m) for m in messages]


def message_from_dict(message: dict) -> Message:
    _type = message["type"]
    if _type == "human":
        return HumanMessage(**message["data"])
    elif _type == "ai":
        return AIMessage(**message["data"])
    elif _type == "system":
        return SystemMessage(**message["data"])
    elif _type == "chat":
        return ChatMessage(**message["data"])
    elif _type == "function":
        return FunctionMessage(**message["data"])
    else:
        raise ValueError(f"Got unexpected message type: {_type}")


def messages_from_dict(messages: list[dict]) -> list[Message]:
    return [message_from_dict(m) for m in messages]



import json
from typing import List, Optional

from pydantic import model_validator, BaseModel, Field, Json

from swarms.structs.agent import Agent
from swarms.structs.task import Task


class Team(BaseModel):
    """
    Class that represents a group of agents, how they should work together and
    their tasks.

    Attributes:
        tasks (Optional[List[Task]]): List of tasks.
        agents (Optional[List[Agent]]): List of agents in this Team.
        architecture (str): Architecture that the Team will follow. Default is "sequential".
        verbose (bool): Verbose mode for the Agent Execution. Default is False.
        config (Optional[Json]): Configuration of the Team. Default is None.
    """

    tasks: Optional[List[Task]] = Field(None, description="List of tasks")
    agents: Optional[List[Agent]] = Field(
        None, description="List of agents in this Team."
    )
    architecture = Field(
        description="architecture that the Team will follow.",
        default="sequential",
    )
    verbose: bool = Field(
        description="Verbose mode for the Agent Execution",
        default=False,
    )
    config: Optional[Json] = Field(
        description="Configuration of the Team.", default=None
    )

    @model_validator(mode="before")
    @classmethod
    def check_config(_cls, values):
        if not values.get("config") and (
            not values.get("agents") and not values.get("tasks")
        ):
            raise ValueError(
                "Either agents and task need to be set or config."
            )

        if values.get("config"):
            config = json.loads(values.get("config"))
            if not config.get("agents") or not config.get("tasks"):
                raise ValueError("Config should have agents and tasks.")

            values["agents"] = [
                Agent(**agent) for agent in config["agents"]
            ]

            tasks = []
            for task in config["tasks"]:
                task_agent = [
                    agt
                    for agt in values["agents"]
                    if agt.role == task["agent"]
                ][0]
                del task["agent"]
                tasks.append(Task(**task, agent=task_agent))

            values["tasks"] = tasks
        return values

    def run(self) -> str:
        """
        Kickoff the Team to work on its tasks.

        Returns:
            output (List[str]): Output of the Team for each task.
        """
        if self.architecture == "sequential":
            return self.__sequential_loop()

    def __sequential_loop(self) -> str:
        """
        Loop that executes the sequential architecture.

        Returns:
            output (str): Output of the Team.
        """
        task_outcome = None
        for task in self.tasks:
            # Add delegation tools to the task if the agent allows it
            # if task.agent.allow_delegation:
            #     tools = AgentTools(agents=self.agents).tools()
            #     task.tools += tools

            self.__log(f"\nWorking Agent: {task.agent.role}")
            self.__log(f"Starting Task: {task.description} ...")

            task_outcome = task.execute(task_outcome)

            self.__log(f"Task output: {task_outcome}")

        return task_outcome

    def __log(self, message):
        if self.verbose:
            print(message)



import concurrent


def execute_concurrently(callable_functions, max_workers=5):
    """
    Executes callable functions concurrently using multithreading.

    Parameters:
    - callable_functions: A list of tuples, each containing the callable function and its arguments.
      For example: [(function1, (arg1, arg2), {'kwarg1': val1}), (function2, (), {})]
    - max_workers: The maximum number of threads to use.

    Returns:
    - results: A list of results returned by the callable functions. If an error occurs in any function,
      the exception object will be placed at the corresponding index in the list.
    """
    results = [None] * len(callable_functions)

    def worker(fn, args, kwargs, index):
        try:
            result = fn(*args, **kwargs)
            results[index] = result
        except Exception as e:
            results[index] = e

    with concurrent.futures.ThreadPoolExecutor(
        max_workers=max_workers
    ) as executor:
        futures = []
        for i, (fn, args, kwargs) in enumerate(callable_functions):
            futures.append(executor.submit(worker, fn, args, kwargs, i))

        # Wait for all threads to complete
        concurrent.futures.wait(futures)

    return results



import inspect


def get_cls_init_params(cls) -> str:
    """
    Get the initialization parameters of a class.

    Args:
        cls: The class to retrieve the initialization parameters from.

    Returns:
        str: A string representation of the initialization parameters.

    """
    init_signature = inspect.signature(cls.__init__)
    params = init_signature.parameters
    params_str_list = []

    for name, param in params.items():
        if name == "self":
            continue
        if name == "kwargs":
            value = "Any keyword arguments"
        elif hasattr(cls, name):
            value = getattr(cls, name)
        else:
            value = cls.__dict__.get(name, "Unknown")

        params_str_list.append(
            f"    {name.capitalize().replace('_', ' ')}: {value}"
        )

    return "\n".join(params_str_list)
